"""Tests for module partial"""

# Author:
#         Laetitia Chapel <laetitia.chapel@irisa.fr>
#
# License: MIT License

import numpy as np
import scipy as sp
import ot
from ot.backend import to_numpy, torch
import pytest


def test_raise_errors():
    n_samples = 20  # nb samples (gaussian)
    n_noise = 20  # nb of samples (noise)

    mu = np.array([0, 0])
    cov = np.array([[1, 0], [0, 2]])

    rng = np.random.RandomState(42)
    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xs = np.append(xs, (rng.rand(n_noise, 2) + 1) * 4).reshape((-1, 2))
    xt = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xt = np.append(xt, (rng.rand(n_noise, 2) + 1) * -3).reshape((-1, 2))

    M = ot.dist(xs, xt)

    p = ot.unif(n_samples + n_noise)
    q = ot.unif(n_samples + n_noise)

    with pytest.raises(ValueError):
        ot.partial.partial_wasserstein_lagrange(p + 1, q, M, 1, log=True)

    with pytest.raises(ValueError):
        ot.partial.partial_wasserstein(p, q, M, m=2, log=True)

    with pytest.raises(ValueError):
        ot.partial.partial_wasserstein(p, q, M, m=-1, log=True)

    with pytest.raises(ValueError):
        ot.partial.entropic_partial_wasserstein(p, q, M, reg=1, m=2, log=True)

    with pytest.raises(ValueError):
        ot.partial.entropic_partial_wasserstein(p, q, M, reg=1, m=-1, log=True)

    with pytest.raises(ValueError):
        ot.partial.partial_gromov_wasserstein(M, M, p, q, m=2, log=True)

    with pytest.raises(ValueError):
        ot.partial.partial_gromov_wasserstein(M, M, p, q, m=-1, log=True)

    with pytest.raises(ValueError):
        ot.partial.entropic_partial_gromov_wasserstein(M, M, p, q, reg=1, m=2, log=True)

    with pytest.raises(ValueError):
        ot.partial.entropic_partial_gromov_wasserstein(
            M, M, p, q, reg=1, m=-1, log=True
        )

    with pytest.raises(AssertionError):
        xs_2d = rng.randn(n_samples, 2)
        xt_2d = rng.randn(n_samples, 2)
        ot.partial.partial_wasserstein_1d(xs_2d, xt_2d)


def test_partial_wasserstein_lagrange():
    n_samples = 20  # nb samples (gaussian)
    n_noise = 20  # nb of samples (noise)

    mu = np.array([0, 0])
    cov = np.array([[1, 0], [0, 2]])

    rng = np.random.RandomState(42)
    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xs = np.append(xs, (rng.rand(n_noise, 2) + 1) * 4).reshape((-1, 2))
    xt = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xt = np.append(xt, (rng.rand(n_noise, 2) + 1) * -3).reshape((-1, 2))

    M = ot.dist(xs, xt)

    p = ot.unif(n_samples + n_noise)
    q = ot.unif(n_samples + n_noise)

    w0, log0 = ot.partial.partial_wasserstein_lagrange(p, q, M, 1, log=True)

    w0, log0 = ot.partial.partial_wasserstein_lagrange(p, q, M, 100, log=True)


def test_partial_wasserstein(nx):
    n_samples = 20  # nb samples (gaussian)
    n_noise = 20  # nb of samples (noise)

    mu = np.array([0, 0])
    cov = np.array([[1, 0], [0, 2]])

    rng = np.random.RandomState(42)
    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xs = np.append(xs, (rng.rand(n_noise, 2) + 1) * 4).reshape((-1, 2))
    xt = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov, random_state=rng)
    xt = np.append(xt, (rng.rand(n_noise, 2) + 1) * -3).reshape((-1, 2))

    M = ot.dist(xs, xt)

    p = ot.unif(n_samples + n_noise)
    q = ot.unif(n_samples + n_noise)

    m = 0.5

    p, q, M = nx.from_numpy(p, q, M)

    w0, log0 = ot.partial.partial_wasserstein(p, q, M, m=m, log=True)
    w, log = ot.partial.entropic_partial_wasserstein(
        p, q, M, reg=1, m=m, log=True, verbose=True
    )

    # check constraints
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=0) - q) <= 1e-5, [True] * len(q))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=0) - q) <= 1e-5, [True] * len(q))

    # check transported mass
    np.testing.assert_allclose(np.sum(to_numpy(w0)), m, atol=1e-04)
    np.testing.assert_allclose(np.sum(to_numpy(w)), m, atol=1e-04)

    w0, log0 = ot.partial.partial_wasserstein2(p, q, M, m=m, log=True)
    w0_val = ot.partial.partial_wasserstein2(p, q, M, m=m, log=False)

    G = log0["T"]

    np.testing.assert_allclose(w0, w0_val, atol=1e-1, rtol=1e-1)

    # check constraints
    np.testing.assert_equal(to_numpy(nx.sum(G, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(G, axis=0) - q) <= 1e-5, [True] * len(q))
    np.testing.assert_allclose(np.sum(to_numpy(G)), m, atol=1e-04)

    empty_array = nx.zeros(0, type_as=M)
    w = ot.partial.partial_wasserstein(empty_array, empty_array, M=M, m=None)

    # check constraints
    np.testing.assert_equal(to_numpy(nx.sum(w, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w, axis=0) - q) <= 1e-5, [True] * len(q))
    np.testing.assert_equal(to_numpy(nx.sum(w, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w, axis=0) - q) <= 1e-5, [True] * len(q))

    # check transported mass
    np.testing.assert_allclose(np.sum(to_numpy(w)), 1, atol=1e-04)

    w0 = ot.partial.entropic_partial_wasserstein(
        empty_array, empty_array, M=M, reg=10, m=None
    )

    # check constraints
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=0) - q) <= 1e-5, [True] * len(q))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=1) - p) <= 1e-5, [True] * len(p))
    np.testing.assert_equal(to_numpy(nx.sum(w0, axis=0) - q) <= 1e-5, [True] * len(q))

    # check transported mass
    np.testing.assert_allclose(np.sum(to_numpy(w0)), 1, atol=1e-04)


def test_partial_wasserstein2_gradient():
    if torch:
        n_samples = 40

        mu = np.array([0, 0])
        cov = np.array([[1, 0], [0, 2]])

        xs = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov)
        xt = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov)

        M = torch.tensor(ot.dist(xs, xt), requires_grad=True, dtype=torch.float64)

        p = torch.tensor(ot.unif(n_samples), dtype=torch.float64)
        q = torch.tensor(ot.unif(n_samples), dtype=torch.float64)

        m = 0.5

        w, log = ot.partial.partial_wasserstein2(p, q, M, m=m, log=True)

        w.backward()

        assert M.grad is not None
        assert M.grad.shape == M.shape


def test_entropic_partial_wasserstein_gradient():
    if torch:
        n_samples = 40

        mu = np.array([0, 0])
        cov = np.array([[1, 0], [0, 2]])

        xs = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov)
        xt = ot.datasets.make_2D_samples_gauss(n_samples, mu, cov)

        M = torch.tensor(ot.dist(xs, xt), requires_grad=True, dtype=torch.float64)

        p = torch.tensor(ot.unif(n_samples), requires_grad=True, dtype=torch.float64)
        q = torch.tensor(ot.unif(n_samples), requires_grad=True, dtype=torch.float64)

        m = 0.5
        reg = 1

        _, log = ot.partial.entropic_partial_wasserstein(
            p, q, M, m=m, reg=reg, log=True
        )

        log["partial_w_dist"].backward()

        assert M.grad is not None
        assert p.grad is not None
        assert q.grad is not None
        assert M.grad.shape == M.shape
        assert p.grad.shape == p.shape
        assert q.grad.shape == q.shape


def test_partial_gromov_wasserstein():
    rng = np.random.RandomState(42)
    n_samples = 20  # nb samples
    n_noise = 10  # nb of samples (noise)

    p = ot.unif(n_samples + n_noise)
    q = ot.unif(n_samples + n_noise)

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    mu_t = np.array([0, 0, 0])
    cov_t = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=rng)
    xs = np.concatenate((xs, ((rng.rand(n_noise, 2) + 1) * 4)), axis=0)
    P = sp.linalg.sqrtm(cov_t)
    xt = rng.randn(n_samples, 3).dot(P) + mu_t
    xt = np.concatenate((xt, ((rng.rand(n_noise, 3) + 1) * 10)), axis=0)
    xt2 = xs[::-1].copy()

    C1 = ot.dist(xs, xs)
    C2 = ot.dist(xt, xt)
    C3 = ot.dist(xt2, xt2)

    m = 2 / 3
    res0, log0 = ot.partial.partial_gromov_wasserstein(
        C1, C3, p, q, m=m, log=True, verbose=True
    )
    np.testing.assert_allclose(res0, 0, atol=1e-1, rtol=1e-1)

    C1 = sp.spatial.distance.cdist(xs, xs)
    C2 = sp.spatial.distance.cdist(xt, xt)

    m = 1
    res0, log0 = ot.partial.partial_gromov_wasserstein(C1, C2, p, q, m=m, log=True)
    G = ot.gromov.gromov_wasserstein(C1, C2, p, q, "square_loss")
    np.testing.assert_allclose(G, res0, atol=1e-04)

    res, log = ot.partial.entropic_partial_gromov_wasserstein(
        C1, C2, p, q, 10, m=m, log=True
    )
    G = ot.gromov.entropic_gromov_wasserstein(C1, C2, p, q, "square_loss", epsilon=10)
    np.testing.assert_allclose(G, res, atol=1e-02)

    w0, log0 = ot.partial.partial_gromov_wasserstein2(C1, C2, p, q, m=m, log=True)
    w0_val = ot.partial.partial_gromov_wasserstein2(C1, C2, p, q, m=m, log=False)
    G = log0["T"]
    np.testing.assert_allclose(w0, w0_val, atol=1e-1, rtol=1e-1)

    m = 2 / 3
    res0, log0 = ot.partial.partial_gromov_wasserstein(C1, C2, p, q, m=m, log=True)
    res, log = ot.partial.entropic_partial_gromov_wasserstein(
        C1, C2, p, q, 100, m=m, log=True
    )

    # check constraints
    np.testing.assert_equal(
        res0.sum(1) <= p, [True] * len(p)
    )  # cf convergence wasserstein
    np.testing.assert_equal(
        res0.sum(0) <= q, [True] * len(q)
    )  # cf convergence wasserstein
    np.testing.assert_allclose(np.sum(res0), m, atol=1e-04)

    np.testing.assert_equal(
        res.sum(1) <= p, [True] * len(p)
    )  # cf convergence wasserstein
    np.testing.assert_equal(
        res.sum(0) <= q, [True] * len(q)
    )  # cf convergence wasserstein
    np.testing.assert_allclose(np.sum(res), m, atol=1e-04)


def test_partial_wasserstein_1d():
    n_samples = 20  # nb samples

    rng = np.random.RandomState(42)
    xs = rng.randn(n_samples, 1)
    xt = rng.randn(n_samples, 1)

    ind_xs_half, ind_xt_half, marginal_costs_half = ot.partial.partial_wasserstein_1d(
        xs, xt, n_transported_samples=n_samples // 2, p=1
    )

    ind_xs, ind_xt, marginal_costs = ot.partial.partial_wasserstein_1d(xs, xt, p=1)

    np.testing.assert_allclose(
        marginal_costs_half, marginal_costs[: n_samples // 2], atol=1e-04
    )
    np.testing.assert_allclose(
        np.sum(np.abs(np.sort(xs[ind_xs_half]) - np.sort(xt[ind_xt_half]))),
        np.sum(marginal_costs_half),
        atol=1e-04,
    )

    n = 20
    x = np.random.rand(n)
    y = np.random.rand(n)

    M = ot.dist(x[:, None], y[:, None], metric="minkowski", p=1)
    indices_x, indices_y, marginal_costs = ot.partial.partial_wasserstein_1d(
        x, y, n_transported_samples=n
    )
    costs = np.cumsum(marginal_costs)

    for i in [1, 5, 10]:
        np.testing.assert_allclose(
            costs[i - 1] / n,
            ot.partial.partial_wasserstein2([], [], M, m=i / n),
            atol=1e-8,
        )

        t = ot.partial.partial_wasserstein([], [], M, m=i / n)
        ind_x, ind_y = np.where(t > 1e-6)

        np.testing.assert_array_equal(np.sort(indices_x[:i]), np.sort(ind_x))
        np.testing.assert_array_equal(np.sort(indices_y[:i]), np.sort(ind_y))



"""
.. warning::
    The list of automatically imported sub-modules is as follows:
    :py:mod:`ot.lp`, :py:mod:`ot.bregman`, :py:mod:`ot.optim`
    :py:mod:`ot.utils`, :py:mod:`ot.datasets`,
    :py:mod:`ot.gromov`, :py:mod:`ot.smooth`
    :py:mod:`ot.stochastic`, :py:mod:`ot.partial`, :py:mod:`ot.regpath`
    , :py:mod:`ot.unbalanced`, :py:mod:`ot.mapping` .
    The following sub-modules are not imported due to additional dependencies:
    - :any:`ot.dr` : depends on :code:`pymanopt` and :code:`autograd`.
    - :any:`ot.plot` : depends on :code:`matplotlib`
"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#         Nicolas Courty <ncourty@irisa.fr>
#
# License: MIT License

# All submodules and packages
from . import lp
from . import bregman
from . import optim
from . import utils
from . import datasets
from . import da
from . import gromov
from . import smooth
from . import stochastic
from . import unbalanced
from . import partial
from . import backend
from . import regpath
from . import weak
from . import factored
from . import solvers
from . import gaussian
from . import lowrank
from . import gmm

# OT functions
from .lp import (
    emd,
    emd2,
    emd_1d,
    emd2_1d,
    wasserstein_1d,
    binary_search_circle,
    wasserstein_circle,
    semidiscrete_wasserstein2_unif_circle,
    linear_circular_ot,
)
from .bregman import sinkhorn, sinkhorn2, barycenter
from .unbalanced import sinkhorn_unbalanced, barycenter_unbalanced, sinkhorn_unbalanced2
from .da import sinkhorn_lpl1_mm
from .sliced import (
    sliced_wasserstein_distance,
    max_sliced_wasserstein_distance,
    sliced_wasserstein_sphere,
    sliced_wasserstein_sphere_unif,
    linear_sliced_wasserstein_sphere,
)
from .gromov import (
    gromov_wasserstein,
    gromov_wasserstein2,
    gromov_barycenters,
    fused_gromov_wasserstein,
    fused_gromov_wasserstein2,
    lowrank_gromov_wasserstein_samples,
)
from .weak import weak_optimal_transport
from .factored import factored_optimal_transport
from .solvers import solve, solve_gromov, solve_sample
from .lowrank import lowrank_sinkhorn

from .batch import solve_batch, solve_sample_batch, solve_gromov_batch, dist_batch

# utils functions
from .utils import dist, unif, tic, toc, toq

__version__ = "0.9.6.post1"

__all__ = [
    "emd",
    "emd2",
    "emd_1d",
    "sinkhorn",
    "sinkhorn2",
    "utils",
    "datasets",
    "bregman",
    "lp",
    "tic",
    "toc",
    "toq",
    "gromov",
    "emd2_1d",
    "wasserstein_1d",
    "backend",
    "gaussian",
    "dist",
    "unif",
    "barycenter",
    "sinkhorn_lpl1_mm",
    "da",
    "optim",
    "sinkhorn_unbalanced",
    "barycenter_unbalanced",
    "sinkhorn_unbalanced2",
    "sliced_wasserstein_distance",
    "sliced_wasserstein_sphere",
    "linear_sliced_wasserstein_sphere",
    "gromov_wasserstein",
    "gromov_wasserstein2",
    "gromov_barycenters",
    "fused_gromov_wasserstein",
    "fused_gromov_wasserstein2",
    "max_sliced_wasserstein_distance",
    "weak_optimal_transport",
    "factored_optimal_transport",
    "solve",
    "solve_gromov",
    "solve_sample",
    "smooth",
    "stochastic",
    "unbalanced",
    "partial",
    "regpath",
    "solvers",
    "weak",
    "factored",
    "lowrank",
    "gmm",
    "binary_search_circle",
    "wasserstein_circle",
    "semidiscrete_wasserstein2_unif_circle",
    "linear_circular_ot",
    "sliced_wasserstein_sphere_unif",
    "lowrank_sinkhorn",
    "lowrank_gromov_wasserstein_samples",
    "solve_batch",
    "solve_gromov_batch",
    "solve_sample_batch",
    "dist_batch",
]



# -*- coding: utf-8 -*-
"""
OT Barycenter Solvers
"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#         Eloi Tanguy <eloi.tanguy@math.cnrs.fr>
#
# License: MIT License

from ..backend import get_backend
from ..utils import dist
from ._network_simplex import emd, emd2

import numpy as np
import scipy as sp
import scipy.sparse as sps

try:
    import cvxopt  # for cvxopt barycenter solver
    from cvxopt import solvers, matrix, spmatrix
except ImportError:
    cvxopt = False


def scipy_sparse_to_spmatrix(A):
    """Efficient conversion from scipy sparse matrix to cvxopt sparse matrix"""
    coo = A.tocoo()
    SP = spmatrix(coo.data.tolist(), coo.row.tolist(), coo.col.tolist(), size=A.shape)
    return SP


def barycenter(A, M, weights=None, verbose=False, log=False, solver="highs-ipm"):
    r"""Compute the Wasserstein barycenter of distributions A

     The function solves the following optimization problem [16]:

    .. math::
       \mathbf{a} = arg\min_\mathbf{a} \sum_i W_{1}(\mathbf{a},\mathbf{a}_i)

    where :

    - :math:`W_1(\cdot,\cdot)` is the Wasserstein distance (see ot.emd.sinkhorn)
    - :math:`\mathbf{a}_i` are training distributions in the columns of matrix :math:`\mathbf{A}`

    The linear program is solved using the interior point solver from scipy.optimize.
    If cvxopt solver if installed it can use cvxopt

    Note that this problem do not scale well (both in memory and computational time).

    Parameters
    ----------
    A : np.ndarray (d,n)
        n training distributions a_i of size d
    M : np.ndarray (d,d)
        loss matrix for OT
    reg : float
        Regularization term >0
    weights : np.ndarray (n,)
        Weights of each histogram a_i on the simplex (barycentric coordinates)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True
    solver : string, optional
        the solver used, default 'interior-point' use the lp solver from
        scipy.optimize. None, or 'glpk' or 'mosek' use the solver from cvxopt.

    Returns
    -------
    a : (d,) ndarray
        Wasserstein barycenter
    log : dict
        log dictionary return only if log==True in parameters


    References
    ----------

    .. [16] Agueh, M., & Carlier, G. (2011). Barycenters in the Wasserstein space. SIAM Journal on Mathematical Analysis, 43(2), 904-924.


    """

    if weights is None:
        weights = np.ones(A.shape[1]) / A.shape[1]
    else:
        assert len(weights) == A.shape[1]

    n_distributions = A.shape[1]
    n = A.shape[0]

    n2 = n * n
    c = np.zeros((0))
    b_eq1 = np.zeros((0))
    for i in range(n_distributions):
        c = np.concatenate((c, M.ravel() * weights[i]))
        b_eq1 = np.concatenate((b_eq1, A[:, i]))
    c = np.concatenate((c, np.zeros(n)))

    lst_idiag1 = [sps.kron(sps.eye(n), np.ones((1, n))) for i in range(n_distributions)]
    #  row constraints
    A_eq1 = sps.hstack(
        (sps.block_diag(lst_idiag1), sps.coo_matrix((n_distributions * n, n)))
    )

    # columns constraints
    lst_idiag2 = []
    lst_eye = []
    for i in range(n_distributions):
        if i == 0:
            lst_idiag2.append(sps.kron(np.ones((1, n)), sps.eye(n)))
            lst_eye.append(-sps.eye(n))
        else:
            lst_idiag2.append(sps.kron(np.ones((1, n)), sps.eye(n - 1, n)))
            lst_eye.append(-sps.eye(n - 1, n))

    A_eq2 = sps.hstack((sps.block_diag(lst_idiag2), sps.vstack(lst_eye)))
    b_eq2 = np.zeros((A_eq2.shape[0]))

    # full problem
    A_eq = sps.vstack((A_eq1, A_eq2))
    b_eq = np.concatenate((b_eq1, b_eq2))

    if not cvxopt or solver in ["interior-point", "highs", "highs-ipm", "highs-ds"]:
        # cvxopt not installed or interior point

        if solver is None:
            solver = "interior-point"

        options = {"disp": verbose}
        sol = sp.optimize.linprog(
            c, A_eq=A_eq, b_eq=b_eq, method=solver, options=options
        )
        x = sol.x
        b = x[-n:]

    else:
        h = np.zeros((n_distributions * n2 + n))
        G = -sps.eye(n_distributions * n2 + n)

        sol = solvers.lp(
            matrix(c),
            scipy_sparse_to_spmatrix(G),
            matrix(h),
            A=scipy_sparse_to_spmatrix(A_eq),
            b=matrix(b_eq),
            solver=solver,
        )

        x = np.array(sol["x"])
        b = x[-n:].ravel()

    if log:
        return b, sol
    else:
        return b


def free_support_barycenter(
    measures_locations,
    measures_weights,
    X_init,
    b=None,
    weights=None,
    numItermax=100,
    stopThr=1e-7,
    verbose=False,
    log=None,
    numThreads=1,
):
    r"""
    Solves the free support (locations of the barycenters are optimized, not the weights) Wasserstein barycenter problem (i.e. the weighted Frechet mean for the 2-Wasserstein distance), formally:

    .. math::
        \min_\mathbf{X} \quad \sum_{i=1}^N w_i W_2^2(\mathbf{b}, \mathbf{X}, \mathbf{a}_i, \mathbf{X}_i)

    where :

    - :math:`w \in \mathbb{(0, 1)}^{N}`'s are the barycenter weights and sum to one
    - `measure_weights` denotes the :math:`\mathbf{a}_i \in \mathbb{R}^{k_i}`: empirical measures weights (on simplex)
    - `measures_locations` denotes the :math:`\mathbf{X}_i \in \mathbb{R}^{k_i, d}`: empirical measures atoms locations
    - :math:`\mathbf{b} \in \mathbb{R}^{k}` is the desired weights vector of the barycenter

    This problem is considered in :ref:`[20] <references-free-support-barycenter>` (Algorithm 2).
    There are two differences with the following codes:

    - we do not optimize over the weights
    - we do not do line search for the locations updates, we use i.e. :math:`\theta = 1` in
      :ref:`[20] <references-free-support-barycenter>` (Algorithm 2). This can be seen as a discrete
      implementation of the fixed-point algorithm of
      :ref:`[43] <references-free-support-barycenter>` proposed in the continuous setting.

    Parameters
    ----------
    measures_locations : list of N (k_i,d) array-like
        The discrete support of a measure supported on :math:`k_i` locations of a `d`-dimensional space
        (:math:`k_i` can be different for each element of the list)
    measures_weights : list of N (k_i,) array-like
        Numpy arrays where each numpy array has :math:`k_i` non-negatives values summing to one
        representing the weights of each discrete input measure
    X_init : (k,d) array-like
        Initialization of the support locations (on `k` atoms) of the barycenter
    b : (k,) array-like
        Initialization of the weights of the barycenter (non-negatives, sum to 1)
    weights : (N,) array-like
        Initialization of the coefficients of the barycenter (non-negatives, sum to 1)
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True
    numThreads: int or "max", optional (default=1, i.e. OpenMP is not used)
        If compiled with OpenMP, chooses the number of threads to parallelize.
        "max" selects the highest number possible.

    Returns
    -------
    X : (k,d) array-like
        Support locations (on k atoms) of the barycenter

    .. _references-free-support-barycenter:
    References
    ----------
    .. [20] Cuturi, Marco, and Arnaud Doucet. "Fast computation of Wasserstein barycenters."
        International Conference on Machine Learning. 2014.

    .. [43] Álvarez-Esteban, Pedro C., et al. "A fixed-point approach to barycenters in Wasserstein space."
        Journal of Mathematical Analysis and Applications 441.2 (2016): 744-762.

    """

    nx = get_backend(*measures_locations, *measures_weights, X_init)

    iter_count = 0

    N = len(measures_locations)
    k = X_init.shape[0]
    d = X_init.shape[1]
    if b is None:
        b = nx.ones((k,), type_as=X_init) / k
    if weights is None:
        weights = nx.ones((N,), type_as=X_init) / N

    X = X_init

    log_dict = {}
    displacement_square_norms = []

    displacement_square_norm = stopThr + 1.0

    while displacement_square_norm > stopThr and iter_count < numItermax:
        T_sum = nx.zeros((k, d), type_as=X_init)

        for measure_locations_i, measure_weights_i, weight_i in zip(
            measures_locations, measures_weights, weights
        ):
            M_i = dist(X, measure_locations_i)
            T_i = emd(b, measure_weights_i, M_i, numThreads=numThreads)
            T_sum = T_sum + weight_i * 1.0 / b[:, None] * nx.dot(
                T_i, measure_locations_i
            )

        displacement_square_norm = nx.sum((T_sum - X) ** 2)
        if log:
            displacement_square_norms.append(displacement_square_norm)

        X = T_sum

        if verbose:
            print(
                "iteration %d, displacement_square_norm=%f\n",
                iter_count,
                displacement_square_norm,
            )

        iter_count += 1

    if log:
        log_dict["displacement_square_norms"] = displacement_square_norms
        return X, log_dict
    else:
        return X


def generalized_free_support_barycenter(
    X_list,
    a_list,
    P_list,
    n_samples_bary,
    Y_init=None,
    b=None,
    weights=None,
    numItermax=100,
    stopThr=1e-7,
    verbose=False,
    log=None,
    numThreads=1,
    eps=0,
):
    r"""
    Solves the free support generalized Wasserstein barycenter problem: finding a barycenter (a discrete measure with
    a fixed amount of points of uniform weights) whose respective projections fit the input measures.
    More formally:

    .. math::
        \min_\gamma \quad \sum_{i=1}^p w_i W_2^2(\nu_i, \mathbf{P}_i\#\gamma)

    where :

    - :math:`\gamma = \sum_{l=1}^n b_l\delta_{y_l}` is the desired barycenter with each :math:`y_l \in \mathbb{R}^d`
    - :math:`\mathbf{b} \in \mathbb{R}^{n}` is the desired weights vector of the barycenter
    - The input measures are :math:`\nu_i = \sum_{j=1}^{k_i}a_{i,j}\delta_{x_{i,j}}`
    - The :math:`\mathbf{a}_i \in \mathbb{R}^{k_i}` are the respective empirical measures weights (on the simplex)
    - The :math:`\mathbf{X}_i \in \mathbb{R}^{k_i, d_i}` are the respective empirical measures atoms locations
    - :math:`w = (w_1, \cdots w_p)` are the barycenter coefficients (on the simplex)
    - Each :math:`\mathbf{P}_i \in \mathbb{R}^{d, d_i}`, and :math:`P_i\#\nu_i = \sum_{j=1}^{k_i}a_{i,j}\delta_{P_ix_{i,j}}`

    As show by :ref:`[42] <references-generalized-free-support-barycenter>`,
    this problem can be re-written as a Wasserstein Barycenter problem,
    which we solve using the free support method :ref:`[20] <references-generalized-free-support-barycenter>`
    (Algorithm 2).

    Parameters
    ----------
    X_list : list of p (k_i,d_i) array-like
        Discrete supports of the input measures: each consists of :math:`k_i` locations of a `d_i`-dimensional space
        (:math:`k_i` can be different for each element of the list)
    a_list : list of p (k_i,) array-like
        Measure weights: each element is a vector (k_i) on the simplex
    P_list : list of p (d_i,d) array-like
        Each :math:`P_i` is a linear map :math:`\mathbb{R}^{d} \rightarrow \mathbb{R}^{d_i}`
    n_samples_bary : int
        Number of barycenter points
    Y_init : (n_samples_bary,d) array-like
        Initialization of the support locations (on `k` atoms) of the barycenter
    b : (n_samples_bary,) array-like
        Initialization of the weights of the barycenter measure (on the simplex)
    weights : (p,) array-like
        Initialization of the coefficients of the barycenter (on the simplex)
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True
    numThreads: int or "max", optional (default=1, i.e. OpenMP is not used)
        If compiled with OpenMP, chooses the number of threads to parallelize.
        "max" selects the highest number possible.
    eps: Stability coefficient for the change of variable matrix inversion
        If the :math:`\mathbf{P}_i^T` matrices don't span :math:`\mathbb{R}^d`, the problem is ill-defined and a matrix
        inversion will fail. In this case one may set eps=1e-8 and get a solution anyway (which may make little sense)


    Returns
    -------
    Y : (n_samples_bary,d) array-like
        Support locations (on n_samples_bary atoms) of the barycenter


    .. _references-generalized-free-support-barycenter:
    References
    ----------
    .. [20] Cuturi, M. and Doucet, A.. "Fast computation of Wasserstein barycenters."
        International Conference on Machine Learning. 2014.

    .. [42] Delon, J., Gozlan, N., and Saint-Dizier, A.. Generalized Wasserstein barycenters
        between probability measures living on different subspaces.
        arXiv preprint arXiv:2105.09755, 2021.

    """
    nx = get_backend(*X_list, *a_list, *P_list)
    d = P_list[0].shape[1]
    p = len(X_list)

    if weights is None:
        weights = nx.ones(p, type_as=X_list[0]) / p

    # variable change matrix to reduce the problem to a Wasserstein Barycenter (WB)
    A = eps * nx.eye(
        d, type_as=X_list[0]
    )  # if eps nonzero: will force the invertibility of A
    for P_i, lambda_i in zip(P_list, weights):
        A = A + lambda_i * P_i.T @ P_i
    B = nx.inv(nx.sqrtm(A))

    Z_list = [
        x @ Pi @ B.T for (x, Pi) in zip(X_list, P_list)
    ]  # change of variables -> (WB) problem on Z

    if Y_init is None:
        Y_init = nx.randn(n_samples_bary, d, type_as=X_list[0])

    if b is None:
        b = nx.ones(n_samples_bary, type_as=X_list[0]) / n_samples_bary  # not optimized

    out = free_support_barycenter(
        Z_list,
        a_list,
        Y_init,
        b,
        numItermax=numItermax,
        stopThr=stopThr,
        verbose=verbose,
        log=log,
        numThreads=numThreads,
    )

    if log:  # unpack
        Y, log_dict = out
    else:
        Y = out
        log_dict = None
    Y = Y @ B.T  # return to the Generalized WB formulation

    if log:
        return Y, log_dict
    else:
        return Y


def ot_barycenter_energy(measure_locations, measure_weights, X, a, cost_list, nx=None):
    r"""
    Computes the energy of the OT barycenter functional for a given barycenter
    support `X` and weights `a`: .. math::
        V(X, a) = \sum_{k=1}^K w_k \mathcal{T}_{c_k}(X, a, Y_k, b_k),

    where: - :math:`X` (n, d) is the barycenter support, - :math:`a` (n) is the
    barycenter weights, - :math:`Y_k` (m_k, d_k) is the k-th measure support
      (`measure_locations[k]`),
    - :math:`b_k` (m_k) is the k-th measure weights (`measure_weights[k]`),
    - :math:`c_k: \mathbb{R}^{n\times d}\times\mathbb{R}^{m_k\times d_k}
         \rightarrow \mathbb{R}_+^{n\times m_k}` is the k-th cost function
         (which computes the pairwise cost matrix)
    - :math:`\mathcal{T}_{c_k}(X, a, Y_k, b)` is the OT cost between the
      barycenter measure and the k-th measure with respect to the cost
      :math:`c_k`.
    The function computes :math:`V(X, a)` as defined above.

    Parameters
    ----------
    measure_locations : list of array-like
        List of K arrays of measure positions, each of shape (m_k, d_k).
    measure_weights : list of array-like
        List of K arrays of measure weights, each of shape (m_k).
    X : array-like
        Array of shape (n, d) representing barycenter points.
    a : array-like
        Array of shape (n,) representing barycenter weights.
    cost_list : list of callable or callable
        List of K cost functions :math:`c_k: \mathbb{R}^{n\times
        d}\times\mathbb{R}^{m_k\times d_k} \rightarrow \mathbb{R}_+^{n\times
        m_k}`. If cost_list is a single callable, the same cost is used K times.
    nx : backend, optional
        The backend to use.

    Returns
    -------
    V : float
        The value of the OT barycenter functional :math:`V(X, a)`.

    References
    ----------
    .. [77] Tanguy, Eloi and Delon, Julie and Gozlan, Nathaël (2024). Computing
        barycenters of Measures for Generic Transport Costs. arXiv preprint
        2501.04016 (2024)

    See Also
    --------
    ot.lp.free_support_barycenter_generic_costs : Free support solver for the
    associated barycenter problem.
    """
    if nx is None:
        nx = get_backend(*measure_locations, *measure_weights, X, a)
    K = len(measure_locations)
    if callable(cost_list):
        cost_list = [cost_list] * K
    V = 0
    for k in range(K):
        C_k = cost_list[k](X, measure_locations[k])
        V += emd2(a, measure_weights[k], C_k)
    return V


def free_support_barycenter_generic_costs(
    measure_locations,
    measure_weights,
    X_init,
    cost_list,
    ground_bary=None,
    a=None,
    numItermax=100,
    method="L2_barycentric_proj",
    stopThr=1e-5,
    log=False,
    ground_bary_lr=1e-2,
    ground_bary_numItermax=100,
    ground_bary_stopThr=1e-5,
    ground_bary_solver="SGD",
    clean_measure=False,
):
    r"""
    Solves the OT barycenter problem [77] for generic costs using the fixed
    point algorithm, iterating the ground barycenter function B on transport
    plans between the current barycenter and the measures.

    The problem finds an optimal barycenter support `X` of given size (n, d)
    (enforced by the initialisation), minimising a sum of pairwise transport
    costs for the costs :math:`c_k`:

    .. math::
        \min_{X} \sum_{k=1}^K \mathcal{T}_{c_k}(X, a, Y_k, b_k),

    where:

    - :math:`X` (n, d) is the barycenter support,
    - :math:`a` (n) is the (fixed) barycenter weights,
    - :math:`Y_k` (m_k, d_k) is the k-th measure support
      (`measure_locations[k]`),
    - :math:`b_k` (m_k) is the k-th measure weights (`measure_weights[k]`),
    - :math:`c_k: \mathbb{R}^{n\times d}\times\mathbb{R}^{m_k\times d_k}\rightarrow \mathbb{R}_+^{n\times m_k}` is the k-th cost function (which computes the pairwise cost matrix)
    - :math:`\mathcal{T}_{c_k}(X, a, Y_k, b)` is the OT cost between the barycenter measure and the k-th measure with respect to the cost :math:`c_k`:

    .. math::
        \mathcal{T}_{c_k}(X, a, Y_k, b_k) = \min_\pi \quad \langle \pi, c_k(X, Y_k) \rangle_F

        s.t. \ \pi \mathbf{1} = \mathbf{a}

             \pi^T \mathbf{1} = \mathbf{b_k}

             \pi \geq 0

    in other words, :math:`\mathcal{T}_{c_k}(X, a, Y_k, b_k)` is `ot.emd2(a, b_k, c_k(X, Y_k))`.

    The function :math:`B:\mathbb{R}^{n\times d_1}\times
    \cdots\times\mathbb{R}^{n\times d_K} \longrightarrow \mathbb{R}^{n\times d}`
    is an input to the solver. `B` computes solutions of the following
    minimisation problem given :math:`(Y_1, \cdots, Y_K) \in \mathbb{R}^{n\times
    d_1}\times\cdots\times\mathbb{R}^{n\times d_K}` (broadcasted along `n`):

    .. math::
        B(y_1, \cdots, y_K) = \mathrm{argmin}_{x \in \mathbb{R}^d} \sum_{k=1}^K c_k(x, y_k),

    The input function B takes a list of K arrays of shape (n, d_k) and returns
    an array of shape (n, d). For certain costs, :math:`B` can be computed
    explicitly, or through a numerical solver.

    This function implements two algorithms:

    - Algorithm 2 from [77] when `method=true_fixed_point` is used, which may
      increase the support size of the barycenter at each iteration, with a
      maximum final size of :math:`N_0 + T\sum_k n_k - TK` for T iterations and
      an initial support size of :math:`N_0`. The computation of the iterates is
      done using the North West Corner multi-marginal gluing method. This method
      has convergence guarantees [77].

    - Algorithm 3 from [77] when `method=L2_barycentric_proj` is used, which is
      a heuristic simplification which fixes the weights and support size of the
      barycenter by performing barycentric projections of the pair-wise OT
      matrices. This method is substantially faster than the first one, but does
      not have convergence guarantees. (Default)

    The implemented methods ([77] Algorithms 2 and 3), generalises [20] and [43]
    to general costs and includes convergence guarantees, including for discrete
    measures.

    Parameters
    ----------
    measure_locations : list of array-like
        List of K arrays of measure positions, each of shape (m_k, d_k).
    measure_weights : list of array-like
        List of K arrays of measure weights, each of shape (m_k).
    X_init : array-like
        Array of shape (n, d) representing initial barycenter points.
    cost_list : list of callable or callable
        List of K cost functions :math:`c_k: \mathbb{R}^{n\times
        d}\times\mathbb{R}^{m_k\times d_k} \rightarrow \mathbb{R}_+^{n\times
        m_k}`. If cost_list is a single callable, the same cost is used K times.
    ground_bary : callable or None, optional
        Function List(array(n, d_k)) -> array(n, d) accepting a list of K arrays
        of shape :math:`(n\times d_K)`, computing the ground barycenters
        (broadcasted over n). If not provided, done with Adam on PyTorch
        (requires PyTorch backend), inefficiently using the cost functions in
        `cost_list`.
    a : array-like, optional
        Array of shape (n,) representing weights of the barycenter
        measure.Defaults to uniform.
    numItermax : int, optional
        Maximum number of iterations (default is 100).
    method : str, optional
        Barycentre method: 'L2_barycentric_proj' (default) for Euclidean
        barycentric projection, or 'true_fixed_point' for iterates using the
        North West Corner multi-marginal gluing method.
    stopThr : float, optional
        If :math:`W_2^2(a_t, X_t, a_{t+1}, X_{t+1}) < \mathrm{stopThr} \times
        \frac{1}{n}\|X_t\|_2^2`, terminate (default is 1e-5).
    log : bool, optional
        Whether to return the log dictionary (default is False).
    ground_bary_lr : float, optional
        Learning rate for the ground barycenter solver (if auto is used).
    ground_bary_numItermax : int, optional
        Maximum number of iterations for the ground barycenter solver (if auto
        is used).
    ground_bary_stopThr : float, optional
        Stop threshold for the ground barycenter solver (if auto is used): stop
        if the energy decreases less than this value.
    ground_bary_solver : str, optional
        Solver for auto ground bary solver (torch SGD or Adam). Default is
        "SGD".
    clean_measure : bool, optional
        For method=='true_fixed_point', whether to clean the discrete measure
        (X, a) at each iteration to remove duplicate points and sum their
        weights (default is False).

    Returns
    -------
    X : array-like
        Array of shape (n, d) representing barycenter points.
    log_dict : list of array-like, optional
        log containing the exit status, list of iterations and list of
        displacements if log is True.

    References
    ----------
    .. [77] Tanguy, Eloi and Delon, Julie and Gozlan, Nathaël (2024). Computing
        barycenters of Measures for Generic Transport Costs. arXiv preprint
        2501.04016 (2024)

    .. [20] Cuturi, Marco, and Arnaud Doucet. "Fast computation of Wasserstein
        barycenters." International Conference on Machine Learning. 2014.

    .. [43] Álvarez-Esteban, Pedro C., et al. "A fixed-point approach to
        barycenters in Wasserstein space." Journal of Mathematical Analysis and
        Applications 441.2 (2016): 744-762.

    .. note:: For the case of the L2 cost :math:`c_k(x, y) = \|x-y\|_2^2`,
        the ground barycenter is simply the Euclidean barycenter, i.e.
        :math:`B(y_1, \cdots, y_K) = \sum_k w_k y_k`. In this case, we recover
        the free-support algorithm from [20].

    See Also
    --------
    ot.lp.free_support_barycenter : Free support solver for the case where
    :math:`c_k(x,y) = \lambda_k\|x-y\|_2^2`.

    ot.lp.generalized_free_support_barycenter : Free support solver for the case
    where :math:`c_k(x,y) = \|P_kx-y\|_2^2` with :math:`P_k` linear.

    ot.lp.NorthWestMMGluing : gluing method used in the `true_fixed_point`
    method.
    """
    assert method in [
        "L2_barycentric_proj",
        "true_fixed_point",
    ], "Method must be 'L2_barycentric_proj' or 'true_fixed_point'"
    nx = get_backend(X_init, measure_locations[0])
    K = len(measure_locations)
    n = X_init.shape[0]
    if a is None:
        a = nx.ones(n, type_as=X_init) / n
    if callable(cost_list):  # use the given cost for all K pairs
        cost_list = [cost_list] * K
    auto_ground_bary = False

    if ground_bary is None:
        auto_ground_bary = True
        assert str(nx) == "torch", (
            f"Backend {str(nx)} is not compatible with ground_bary=None, it"
            "must be provided if not using PyTorch backend"
        )
        try:
            import torch
            from torch.optim import Adam, SGD

            def ground_bary(y, x_init):
                x = x_init.clone().detach().requires_grad_(True)
                solver = Adam if ground_bary_solver == "Adam" else SGD
                opt = solver([x], lr=ground_bary_lr)
                loss_prev = None
                for i in range(ground_bary_numItermax):
                    opt.zero_grad()
                    # inefficient cost computation but compatible
                    # with the choice of cost_list[k] giving the cost matrix
                    loss = torch.sum(
                        torch.stack(
                            [torch.diag(cost_list[k](x, y[k])) for k in range(K)]
                        )
                    )
                    loss.backward()
                    opt.step()
                    if i == 0:
                        diff = ground_bary_stopThr + 1.0
                    else:
                        diff = loss_prev - loss.item()
                    loss_prev = loss.item()
                    if diff < ground_bary_stopThr:
                        break
                return x.detach()

        except ImportError:
            raise ImportError("PyTorch is required to use ground_bary=None")

    X_list = [X_init] if log else []  # store the iterations
    a_list = [a] if log and method == "true_fixed_point" else []
    X = X_init
    diff_list = []  # store energy differences
    V_list = []  # store energy values
    exit_status = "Max iterations reached"

    for i in range(numItermax):
        pi_list = [  # compute the pairwise transport plans
            emd(a, measure_weights[k], cost_list[k](X, measure_locations[k]))
            for k in range(K)
        ]
        if i == 0:  # compute initial energy
            V = ot_barycenter_energy(
                measure_locations, measure_weights, X, a, cost_list, nx=nx
            )
            if log:
                V_list.append(V)

        if method == "L2_barycentric_proj":
            a_next = a  # barycentre weights are fixed
            Y_perm = [
                (1 / a[:, None]) * pi_list[k] @ measure_locations[k] for k in range(K)
            ]  # L2 barycentric projection of pi_k
            if auto_ground_bary:  # use previous position as initialization
                X_next = ground_bary(Y_perm, X)
            else:
                X_next = ground_bary(Y_perm)

        elif method == "true_fixed_point":
            # North West Corner gluing of pi_k
            J, a_next = NorthWestMMGluing(pi_list, nx=nx)
            # J is a (N, K) array of indices, w is a (N,) array of weights
            # Each Y_perm[k] is a (N, d_k) array of some points in Y_list[k]
            Y_perm = [measure_locations[k][J[:, k]] for k in range(K)]
            # warm start impossible due to possible size mismatch
            X_next = ground_bary(Y_perm)

            if clean_measure and method == "true_fixed_point":
                # clean the discrete measure (X, a) to remove duplicates
                X_next, a_next = _clean_discrete_measure(X_next, a_next, nx=nx)

        V_next = ot_barycenter_energy(
            measure_locations, measure_weights, X_next, a_next, cost_list, nx=nx
        )
        diff = V - V_next
        if log:
            X_list.append(X_next)
            V_list.append(V_next)
            if method == "true_fixed_point":
                a_list.append(a_next)
            diff_list.append(diff)

        X = X_next
        a = a_next
        V = V_next

        if diff < stopThr:
            exit_status = "Stationary Point"
            break

    if log:
        log_dict = {
            "X_list": X_list,
            "exit_status": exit_status,
            "a_list": a_list,
            "diff_list": diff_list,
            "V_list": V_list,
        }
        if method == "true_fixed_point":
            return X, a, log_dict
        else:
            return X, log_dict

    if method == "true_fixed_point":
        return X, a
    else:
        return X


def _to_int_array(x, nx=None):
    """
    Converts an array to an integer type array.
    """
    if nx is None:
        nx = get_backend(x)
    if str(nx) == "numpy":
        return x.astype(int)

    if str(nx) == "torch":
        return x.to(int)

    if str(nx) == "jax":
        return x.astype(int)

    if str(nx) == "cupy":
        return x.astype(int)

    if str(nx) == "tf":
        import tensorflow as tf

        return tf.cast(x, tf.int32)


def NorthWestMMGluing(pi_list, a=None, log=False, nx=None):
    r"""
    Glue transport plans :math:`(\pi_1, ..., \pi_K)` which have a common first
    marginal using the (multi-marginal) North-West Corner method. Writing the
    marginals of each :math:`\pi_k\in \mathbb{R}^{n\times n_l}` as :math:`a \in
    \mathbb{R}^n` and :math:`b_k \in \mathbb{R}^{n_k}`, the output represents a
    particular K-marginal transport plan :math:`\rho \in
    \mathbb{R}^{n_1\times\cdots\times n_K}` whose k-th marginal is :math:`b_k`.
    This K-plan is such that there exists a K+1-marginal transport plan
    :math:`\gamma \in \mathbb{R}^{n\times n_1 \times \cdots \times n_K}` such
    that :math:`\sum_i\gamma_{i,j_1,\cdots,j_K} = \rho_{j_1, \cdots, j_K}` and
    with Einstein summation convention, :math:`\gamma_{i, j_1, \cdots, j_K} =
    [\pi_k]_{i, j_k}` for all :math:`k=1,\cdots,K`.

    Instead of outputting the full K-multi-marginal plan :math:`\rho`, this
    function provides an array `J` of shape (N, K) where each `J[i]` is of the
    form `(J[i, 1], ..., J[i, K])` with each `J[i, k]` between 0 and
    :math:`n_k-1`, and a weight vector `w` of size N, such that the K-plan
    :math:`\rho` writes:

    .. math::
        \rho_{j_1, \cdots, j_K} = \mathbb{1}\left(\exists i \text{ s.t. } (j_1, \cdots, j_K) = (J[i, 1], \cdots, J[i, K])\right)\ w_i.

    This representation is useful for its memory efficiency, as it avoids
    storing the full K-marginal plan.

    If `log=True`, the function computes the full K+1-marginal transport plan
    :math:`\gamma` and stores it in log_dict['gamma']. Note that this option is
    extremely costly in memory.

    Parameters
    ----------
    pi_list : list of arrays (n, n_k)
        List of transport plans.
    log : bool, optional
        If True, return a log dictionary (computationally expensive).
    a : array (n,), optional
        The common first marginal of each transport plan. If None is provided,
        it is computed as the first marginal of the first transport plan.
    nx : backend, optional
        The backend to use. If None is provided, the backend of the first
        transport plan is used.

    Returns
    -------
    J : array (N, K)
        The indices (J[i, 1], ..., J[i, K]) of the K-plan :math:`\rho`.
    w : array (N,)
        The weights w_i of the K-plan :math:`\rho`.
    log_dict : dict, optional
        If log=True, a dictionary containing the full K+1-marginal transport
        plan under the key 'gamma'.
    """
    if nx is None:
        nx = get_backend(pi_list[0])
    if a is None:
        a = nx.sum(pi_list[0], axis=1)  # common first marginal a in Delta_n
    nk_list = [pi.shape[1] for pi in pi_list]  # list of n_k
    K = len(pi_list)
    n = pi_list[0].shape[0]  # number of points in the first marginal
    gamma = None

    log_dict = {}
    if log:  # n x n_1 x ... x n_K tensor
        gamma = nx.zeros([n] + nk_list, type_as=pi_list[0])

    gamma_weights = {}  # dict of (j_1, ..., j_K) : weight
    P_list = [nx.copy(pi) for pi in pi_list]  # copy of the transport plans

    # jjs is a list of K lists of size m_k
    # checks if each jj_idx[k] is < m_k
    # this is to avoid over-shooting the while loop due to numerical
    # imprecision in the conditions "x > 0"
    def jj_idx_in_range(jj_idx, jjs):
        out = True
        for k in range(K):
            out = out and jj_idx[k] < len(jjs[k])
        return out

    for i in range(n):
        # jjs[k] is the list of indices j in [0, n_k - 1] such that Pk[i, j] >0
        jjs = [nx.to_numpy(nx.where(P[i, :] > 0)[0]) for P in P_list]
        # list [0, ..., 0] of size K for use with jjs: current indices in jjs
        jj_idx = [0] * K
        u = a[i]  # mass at i, will decrease to 0 as we fill gamma[i, :]

        # while there is mass to add to gamma[i, :]
        while u > 0 and jj_idx_in_range(jj_idx, jjs):
            # current multi-index j_1 ... j_K
            jj = tuple(jjs[k][jj_idx[k]] for k in range(K))
            # min transport plan value: min_k pi_k[i, j_k]
            v = nx.min(nx.stack([P_list[k][i, jj[k]] for k in range(K)]))
            if log:  # assign mass v to gamma[i, j_1, ..., j_K]
                gamma[(i,) + jj] = v
            if jj in gamma_weights:
                gamma_weights[jj] += v
            else:
                gamma_weights[jj] = v
            u -= v  # at i, we u-v mass left to assign
            for k in range(K):  # update plan copies Pk
                P_list[k][i, jj[k]] -= v  # Pk[i, j_k] has v less mass left
                if P_list[k][i, jj[k]] == 0:
                    # move to next index in jjs[k] if Pk[i, j_k] is empty
                    jj_idx[k] += 1

    log_dict["gamma"] = gamma
    J = list(gamma_weights.keys())  # list of multi-indices (j_1, ..., j_K)
    J = _to_int_array(nx.from_numpy(np.array(J), type_as=pi_list[0]), nx=nx)
    w = nx.stack(list(gamma_weights.values()))
    if log:
        return J, w, log_dict
    return J, w


def _clean_discrete_measure(X, a, tol=1e-10, nx=None):
    r"""
    Simplifies a discrete measure by consolidating duplicate points and summing
    their weights. Given a discrete measure with support X (n, d) and weights a
    (n), returns a points Y (m, d) and weights b (m) such that Y is the set of
    unique points in X and b is the sum of weights in a for each point in Y

    Parameters
    ----------
    X : array-like
        Array of shape (n, d) representing the support points of the discrete
        measure.
    a : array-like
        Array of shape (n,) representing the weights associated with the support
        points.
    tol : float, optional
        Tolerance for determining uniqueness of points in `X`. Points closer
        than `tol` are considered identical. Default is 1e-10.
    nx : backend, optional
        The backend to use. If None is provided, the backend of `X` is used

    Returns
    -------
    Y : array-like
        Array of shape (m, d) representing the unique support points of the
        discrete measure.
    b : array-like
        Array of shape (m,) representing the summed weights for each unique
        point in `Y`.
    """
    nx = get_backend(X, a)
    D = dist(X, X)
    # each D[I[k], J[k]] < tol so X[I[k]] = X[J[k]]
    idxI, idxJ = nx.where(D < tol)
    idxI = nx.to_numpy(idxI)
    idxJ = nx.to_numpy(idxJ)
    # keep only the cases I[k] <= J[k] to avoid pairs (i, j) (j, i) with i != j
    mask = idxI <= idxJ
    idxI, idxJ = idxI[mask], idxJ[mask]
    X_idx_to_Y_idx = {}  # X[i] = Y[X_idx_to_Y_idx[i]]
    # indices of unique points in X, at the end, Y := X[unique_X_idx]
    unique_X_idx = []

    b = []
    for i, j in zip(idxI, idxJ):
        if i not in X_idx_to_Y_idx:  # i is a new point
            unique_X_idx.append(i)
            X_idx_to_Y_idx[i] = len(unique_X_idx) - 1
            b.append(a[i])

        else:  # i is not new, check if j is known
            if j not in X_idx_to_Y_idx:
                b[X_idx_to_Y_idx[i]] += a[j]
                X_idx_to_Y_idx[j] = X_idx_to_Y_idx[i]

    # create the unique points array Y
    Y = X[tuple(unique_X_idx), :]
    b = nx.from_numpy(np.array(b), type_as=X)
    return Y, b



"""Tests for main module ot"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#
# License: MIT License

import warnings

import numpy as np
import pytest

import ot
from ot.datasets import make_1D_gauss as gauss
from ot.backend import torch, tf, get_backend


def test_emd_dimension_and_mass_mismatch():
    # test emd and emd2 for dimension mismatch
    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples + 1)

    M = ot.dist(x, x)

    np.testing.assert_raises(AssertionError, ot.emd, a, a, M)

    np.testing.assert_raises(AssertionError, ot.emd2, a, a, M)

    # test emd and emd2 for mass mismatch
    a = ot.utils.unif(n_samples)
    b = a.copy()
    a[0] = 100
    np.testing.assert_raises(AssertionError, ot.emd, a, b, M)
    np.testing.assert_raises(AssertionError, ot.emd2, a, b, M)


def test_emd_backends(nx):
    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples, n_features)
    y = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples)

    M = ot.dist(x, y)

    G = ot.emd(a, a, M)

    ab, Mb = nx.from_numpy(a, M)

    Gb = ot.emd(ab, ab, Mb)

    np.allclose(G, nx.to_numpy(Gb))


def test_emd2_backends(nx):
    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples, n_features)
    y = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples)

    M = ot.dist(x, y)

    val = ot.emd2(a, a, M)

    ab, Mb = nx.from_numpy(a, M)

    valb = ot.emd2(ab, ab, Mb)

    # check with empty inputs
    valb2 = ot.emd2([], [], Mb)

    np.allclose(val, nx.to_numpy(valb))
    np.allclose(val, nx.to_numpy(valb2))


def test_emd_emd2_types_devices(nx):
    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples, n_features)
    y = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples)

    M = ot.dist(x, y)

    for tp in nx.__type_list__:
        print(nx.dtype_device(tp))

        ab, Mb = nx.from_numpy(a, M, type_as=tp)

        Gb = ot.emd(ab, ab, Mb)

        w = ot.emd2(ab, ab, Mb)

        nx.assert_same_dtype_device(Mb, Gb)
        nx.assert_same_dtype_device(Mb, w)


@pytest.mark.skipif(not tf, reason="tf not installed")
def test_emd_emd2_devices_tf():
    nx = ot.backend.TensorflowBackend()

    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)
    x = rng.randn(n_samples, n_features)
    y = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples)
    M = ot.dist(x, y)

    # Check that everything stays on the CPU
    with tf.device("/CPU:0"):
        ab, Mb = nx.from_numpy(a, M)
        Gb = ot.emd(ab, ab, Mb)
        w = ot.emd2(ab, ab, Mb)
        nx.assert_same_dtype_device(Mb, Gb)
        nx.assert_same_dtype_device(Mb, w)

    if len(tf.config.list_physical_devices("GPU")) > 0:
        # Check that everything happens on the GPU
        ab, Mb = nx.from_numpy(a, M)
        Gb = ot.emd(ab, ab, Mb)
        w = ot.emd2(ab, ab, Mb)
        nx.assert_same_dtype_device(Mb, Gb)
        nx.assert_same_dtype_device(Mb, w)
        assert nx.dtype_device(Gb)[1].startswith("GPU")


def test_emd2_gradients():
    n_samples = 100
    n_features = 2
    rng = np.random.RandomState(0)

    x = rng.randn(n_samples, n_features)
    y = rng.randn(n_samples, n_features)
    a = ot.utils.unif(n_samples)

    M = ot.dist(x, y)

    if torch:
        a1 = torch.tensor(a, requires_grad=True)
        b1 = torch.tensor(a, requires_grad=True)
        M1 = torch.tensor(M, requires_grad=True)

        val, log = ot.emd2(a1, b1, M1, log=True)

        val.backward()

        assert a1.shape == a1.grad.shape
        assert b1.shape == b1.grad.shape
        assert M1.shape == M1.grad.shape

        assert np.allclose(
            a1.grad.cpu().detach().numpy(),
            log["u"].cpu().detach().numpy() - log["u"].cpu().detach().numpy().mean(),
        )

        assert np.allclose(
            b1.grad.cpu().detach().numpy(),
            log["v"].cpu().detach().numpy() - log["v"].cpu().detach().numpy().mean(),
        )

        # Testing for bug #309, checking for scaling of gradient
        a2 = torch.tensor(a, requires_grad=True)
        b2 = torch.tensor(a, requires_grad=True)
        M2 = torch.tensor(M, requires_grad=True)

        val = 10.0 * ot.emd2(a2, b2, M2)

        val.backward()

        assert np.allclose(
            10.0 * a1.grad.cpu().detach().numpy(), a2.grad.cpu().detach().numpy()
        )
        assert np.allclose(
            10.0 * b1.grad.cpu().detach().numpy(), b2.grad.cpu().detach().numpy()
        )
        assert np.allclose(
            10.0 * M1.grad.cpu().detach().numpy(), M2.grad.cpu().detach().numpy()
        )


def test_emd_emd2():
    # test emd and emd2 for simple identity
    n = 100
    rng = np.random.RandomState(0)

    x = rng.randn(n, 2)
    u = ot.utils.unif(n)

    M = ot.dist(x, x)

    G = ot.emd(u, u, M)

    # check G is identity
    np.testing.assert_allclose(G, np.eye(n) / n)
    # check constraints
    np.testing.assert_allclose(u, G.sum(1))  # cf convergence sinkhorn
    np.testing.assert_allclose(u, G.sum(0))  # cf convergence sinkhorn

    w = ot.emd2(u, u, M)
    # check loss=0
    np.testing.assert_allclose(w, 0)


def test_omp_emd2():
    # test emd2 and emd2 with openmp for simple identity
    n = 100
    rng = np.random.RandomState(0)

    x = rng.randn(n, 2)
    u = ot.utils.unif(n)

    M = ot.dist(x, x)

    w = ot.emd2(u, u, M)
    w2 = ot.emd2(u, u, M, numThreads=2)

    np.testing.assert_allclose(w, w2)


def test_emd_empty():
    # test emd and emd2 for simple identity
    n = 100
    rng = np.random.RandomState(0)

    x = rng.randn(n, 2)
    u = ot.utils.unif(n)

    M = ot.dist(x, x)

    G = ot.emd([], [], M)

    # check G is identity
    np.testing.assert_allclose(G, np.eye(n) / n)
    # check constraints
    np.testing.assert_allclose(u, G.sum(1))  # cf convergence sinkhorn
    np.testing.assert_allclose(u, G.sum(0))  # cf convergence sinkhorn

    w = ot.emd2([], [], M)
    # check loss=0
    np.testing.assert_allclose(w, 0)


def test_emd2_multi():
    n = 500  # nb bins

    # bin positions
    x = np.arange(n, dtype=np.float64)

    # Gaussian distributions
    a = gauss(n, m=20, s=5)  # m= mean, s= std

    ls = np.arange(20, 500, 100)
    nb = len(ls)
    b = np.zeros((n, nb))
    for i in range(nb):
        b[:, i] = gauss(n, m=ls[i], s=10)

    # loss matrix
    M = ot.dist(x.reshape((n, 1)), x.reshape((n, 1)))
    # M/=M.max()

    print("Computing {} EMD ".format(nb))

    # emd loss 1 proc
    ot.tic()
    emd1 = ot.emd2(a, b, M, 1)
    ot.toc("1 proc : {} s")

    # emd loss multipro proc
    ot.tic()
    emdn = ot.emd2(a, b, M)
    ot.toc("multi proc : {} s")

    np.testing.assert_allclose(emd1, emdn)

    # emd loss multipro proc with log
    ot.tic()
    emdn = ot.emd2(a, b, M, log=True, return_matrix=True)
    ot.toc("multi proc : {} s")

    for i in range(len(emdn)):
        emd = emdn[i]
        log = emd[1]
        cost = emd[0]
        check_duality_gap(a, b[:, i], M, log["G"], log["u"], log["v"], cost)
        emdn[i] = cost

    emdn = np.array(emdn)
    np.testing.assert_allclose(emd1, emdn)


def test_lp_barycenter():
    a1 = np.array([1.0, 0, 0])[:, None]
    a2 = np.array([0, 0, 1.0])[:, None]

    A = np.hstack((a1, a2))
    M = np.array([[0, 1.0, 4.0], [1.0, 0, 1.0], [4.0, 1.0, 0]])

    # obvious barycenter between two Diracs
    bary0 = np.array([0, 1.0, 0])

    bary = ot.lp.barycenter(A, M, [0.5, 0.5])

    np.testing.assert_allclose(bary, bary0, rtol=1e-5, atol=1e-7)
    np.testing.assert_allclose(bary.sum(), 1)


def test_free_support_barycenter():
    measures_locations = [
        np.array([-1.0]).reshape((1, 1)),
        np.array([1.0]).reshape((1, 1)),
    ]
    measures_weights = [np.array([1.0]), np.array([1.0])]

    X_init = np.array([-12.0]).reshape((1, 1))

    # obvious barycenter location between two Diracs
    bar_locations = np.array([0.0]).reshape((1, 1))

    X = ot.lp.free_support_barycenter(measures_locations, measures_weights, X_init)

    np.testing.assert_allclose(X, bar_locations, rtol=1e-5, atol=1e-7)


def test_free_support_barycenter_backends(nx):
    measures_locations = [
        np.array([-1.0]).reshape((1, 1)),
        np.array([1.0]).reshape((1, 1)),
    ]
    measures_weights = [np.array([1.0]), np.array([1.0])]
    X_init = np.array([-12.0]).reshape((1, 1))

    X = ot.lp.free_support_barycenter(measures_locations, measures_weights, X_init)

    measures_locations2 = nx.from_numpy(*measures_locations)
    measures_weights2 = nx.from_numpy(*measures_weights)
    X_init2 = nx.from_numpy(X_init)

    X2 = ot.lp.free_support_barycenter(measures_locations2, measures_weights2, X_init2)

    np.testing.assert_allclose(X, nx.to_numpy(X2))


def test_generalised_free_support_barycenter():
    X = [
        np.array([-1.0, -1.0]).reshape((1, 2)),
        np.array([1.0, 1.0]).reshape((1, 2)),
    ]  # two 2D points bar is obviously 0
    a = [np.array([1.0]), np.array([1.0])]

    P = [np.eye(2), np.eye(2)]

    Y_init = np.array([-12.0, 7.0]).reshape((1, 2))

    # obvious barycenter location between two 2D Diracs
    Y_true = np.array([0.0, 0.0]).reshape((1, 2))

    # test without log and no init
    Y = ot.lp.generalized_free_support_barycenter(X, a, P, 1)
    np.testing.assert_allclose(Y, Y_true, rtol=1e-5, atol=1e-7)

    # test with log and init
    Y, _ = ot.lp.generalized_free_support_barycenter(
        X, a, P, 1, Y_init=Y_init, b=np.array([1.0]), log=True
    )
    np.testing.assert_allclose(Y, Y_true, rtol=1e-5, atol=1e-7)


def test_generalised_free_support_barycenter_backends(nx):
    X = [np.array([-1.0]).reshape((1, 1)), np.array([1.0]).reshape((1, 1))]
    a = [np.array([1.0]), np.array([1.0])]
    P = [np.array([1.0]).reshape((1, 1)), np.array([1.0]).reshape((1, 1))]
    Y_init = np.array([-12.0]).reshape((1, 1))

    Y = ot.lp.generalized_free_support_barycenter(X, a, P, 1, Y_init=Y_init)

    X2 = nx.from_numpy(*X)
    a2 = nx.from_numpy(*a)
    P2 = nx.from_numpy(*P)
    Y_init2 = nx.from_numpy(Y_init)

    Y2 = ot.lp.generalized_free_support_barycenter(X2, a2, P2, 1, Y_init=Y_init2)

    np.testing.assert_allclose(Y, nx.to_numpy(Y2))


def test_free_support_barycenter_generic_costs():
    measures_locations = [
        np.array([-1.0]).reshape((1, 1)),
        np.array([1.0]).reshape((1, 1)),
    ]
    measures_weights = [np.array([1.0]), np.array([1.0])]

    X_init = np.array([-12.0]).reshape((1, 1))

    # obvious barycenter location between two Diracs
    bar_locations = np.array([0.0]).reshape((1, 1))

    def cost(x, y):
        return ot.dist(x, y)

    cost_list = [cost, cost]

    def ground_bary(y):
        out = 0
        for yk in y:
            out += yk / len(y)
        return out

    X = ot.lp.free_support_barycenter_generic_costs(
        measures_locations, measures_weights, X_init, cost_list, ground_bary
    )

    np.testing.assert_allclose(X, bar_locations, rtol=1e-5, atol=1e-7)

    # test with log and specific weights
    X2, log = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        a=ot.unif(1),
        log=True,
    )

    assert "X_list" in log
    assert "exit_status" in log
    assert "diff_list" in log

    np.testing.assert_allclose(X, X2, rtol=1e-5, atol=1e-7)

    # test with one iteration for Max Iterations Reached
    X3, log2 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        numItermax=1,
        log=True,
    )
    assert log2["exit_status"] == "Max iterations reached"

    # test with a single callable cost
    X3, log3 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost,
        ground_bary,
        numItermax=1,
        log=True,
    )

    # test with no ground_bary but in numpy: requires pytorch backend
    with pytest.raises(AssertionError):
        ot.lp.free_support_barycenter_generic_costs(
            measures_locations,
            measures_weights,
            X_init,
            cost_list,
            ground_bary=None,
            numItermax=1,
        )

    # test with unknown method
    with pytest.raises(AssertionError):
        ot.lp.free_support_barycenter_generic_costs(
            measures_locations,
            measures_weights,
            X_init,
            cost_list,
            ground_bary,
            numItermax=1,
            method="unknown_method",
        )

    # test true fixed-point method
    X4, a4, log4 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        numItermax=3,
        method="true_fixed_point",
        log=True,
    )

    assert "a_list" in log4
    assert X4.shape[0] == a4.shape[0] == 1
    np.testing.assert_allclose(a4, ot.unif(1), rtol=1e-5, atol=1e-7)
    np.testing.assert_allclose(X, X4, rtol=1e-5, atol=1e-7)

    # test with measure cleaning and no log
    X5, a5 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        numItermax=3,
        method="true_fixed_point",
        clean_measure=True,
    )
    np.testing.assert_allclose(a5, ot.unif(1), rtol=1e-5, atol=1e-7)
    np.testing.assert_allclose(X, X5, rtol=1e-5, atol=1e-7)

    # test with (too) lax convergence criterion
    # for Stationary Point exit status
    X6, log6 = ot.lp.free_support_barycenter_generic_costs(
        [np.array([-1.0]).reshape((1, 1))],
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        numItermax=3,
        stopThr=1e20,
        log=True,
    )
    assert log6["exit_status"] == "Stationary Point"


@pytest.mark.skipif(not torch, reason="No torch available")
def test_free_support_barycenter_generic_costs_auto_ground_bary():
    measures_locations = [
        torch.tensor([1.0]).reshape((1, 1)),
        torch.tensor([2.0]).reshape((1, 1)),
    ]
    measures_weights = [torch.tensor([1.0]), torch.tensor([1.0])]

    X_init = torch.tensor([1.2]).reshape((1, 1))

    def cost(x, y):
        return ot.dist(x, y)

    cost_list = [cost, cost]

    def ground_bary(y):
        out = 0
        for yk in y:
            out += yk / len(y)
        return out

    X = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        numItermax=1,
        stopThr=-1,
    )

    X2, log2 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary=None,
        ground_bary_lr=2e-2,
        ground_bary_stopThr=1e-20,
        ground_bary_numItermax=100,
        numItermax=10,
        stopThr=-1,
        log=True,
    )

    np.testing.assert_allclose(X2.numpy(), X.numpy(), rtol=1e-4, atol=1e-4)

    X3 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary=None,
        ground_bary_lr=1e-2,
        ground_bary_stopThr=1e-20,
        ground_bary_numItermax=100,
        numItermax=10,
        ground_bary_solver="Adam",
        stopThr=-1,
    )

    np.testing.assert_allclose(X2.numpy(), X3.numpy(), rtol=1e-3, atol=1e-3)

    # test with (too) lax convergence criterion for ground barycenter
    ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary=None,
        numItermax=1,
        ground_bary_stopThr=100,
    )


@pytest.skip_backend("tf")  # skips because of array assignment
@pytest.skip_backend("jax")
def test_free_support_barycenter_generic_costs_backends(nx):
    measures_locations = [
        np.array([-1.0]).reshape((1, 1)),
        np.array([1.0]).reshape((1, 1)),
    ]
    measures_weights = [np.array([1.0]), np.array([1.0])]
    X_init = np.array([-12.0]).reshape((1, 1))

    def cost(x, y):
        return ot.dist(x, y)

    cost_list = [cost, cost]

    def ground_bary(y):
        out = 0
        for yk in y:
            out += yk / len(y)
        return out

    X = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        method="L2_barycentric_proj",
    )

    measures_locations2 = nx.from_numpy(*measures_locations)
    measures_weights2 = nx.from_numpy(*measures_weights)
    X_init2 = nx.from_numpy(X_init)

    X2 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations2,
        measures_weights2,
        X_init2,
        cost_list,
        ground_bary,
        method="L2_barycentric_proj",
    )

    np.testing.assert_allclose(X, nx.to_numpy(X2))

    X, a = ot.lp.free_support_barycenter_generic_costs(
        measures_locations,
        measures_weights,
        X_init,
        cost_list,
        ground_bary,
        method="true_fixed_point",
    )

    measures_locations2 = nx.from_numpy(*measures_locations)
    measures_weights2 = nx.from_numpy(*measures_weights)
    X_init2 = nx.from_numpy(X_init)

    X2, a2 = ot.lp.free_support_barycenter_generic_costs(
        measures_locations2,
        measures_weights2,
        X_init2,
        cost_list,
        ground_bary,
        method="true_fixed_point",
    )

    np.testing.assert_allclose(a, nx.to_numpy(a2))
    np.testing.assert_allclose(X, nx.to_numpy(X2))

    ot.lp.ot_barycenter_energy(  # test without backend and callable cost
        measures_locations, measures_weights, X, a, cost, nx=None
    )


def verify_gluing_validity(gamma, J, w, pi_list):
    """
    Test the validity of the North-West gluing.
    """
    nx = get_backend(gamma)
    K = len(pi_list)
    n = pi_list[0].shape[0]
    nk_list = [pi.shape[1] for pi in pi_list]

    # Check first marginal
    a = nx.sum(gamma, axis=tuple(range(1, K + 1)))
    assert nx.allclose(a, nx.sum(pi_list[0], axis=1))

    # Check other marginals
    for k in range(K):
        b_k = nx.sum(gamma, axis=tuple(i for i in range(K + 1) if i != k + 1))
        assert nx.allclose(b_k, nx.sum(pi_list[k], axis=0))

    # Check bi-marginals
    for k in range(K):
        gamma_0k = nx.sum(gamma, axis=tuple(i for i in range(1, K + 1) if i != k + 1))
        assert nx.allclose(gamma_0k, pi_list[k])

    # Check that N <= n + sum_k n_k - K
    N = J.shape[0]
    n_k_sum = sum(nk_list)
    assert N <= n + n_k_sum - K, f"N={N}, n={n}, sum(n_k)={n_k_sum}, K={K}"

    # Check that w is on the simplex
    w_sum = nx.sum(w)
    assert nx.allclose(w_sum, 1), f"Sum of weights w is not 1: {w_sum}"

    # Check that gamma_1...K and (J, w) are consistent
    rho = nx.zeros(nk_list, type_as=gamma)
    for i in range(N):
        jj = J[i]
        rho[tuple(jj)] += w[i]

    gamma_1toK = nx.sum(gamma, axis=0)
    assert nx.allclose(rho, gamma_1toK), "rho and gamma_1...K are not consistent"


def test_north_west_mm_gluing():
    rng = np.random.RandomState(0)
    n = 7
    nk_list = [5, 6, 4]
    a = rng.rand(n)
    a = a / np.sum(a)
    b_list = [rng.rand(nk) for nk in nk_list]
    b_list = [b / np.sum(b) for b in b_list]
    M_list = [rng.rand(n, nk) for nk in nk_list]
    pi_list = [ot.emd(a, b, M) for b, M in zip(b_list, M_list)]
    J, w, log_dict = ot.lp.NorthWestMMGluing(pi_list, log=True)
    # Test the validity of the gluing
    gamma = log_dict["gamma"]
    verify_gluing_validity(gamma, J, w, pi_list)

    # test without log
    J2, w2 = ot.lp.NorthWestMMGluing(pi_list, log=False)
    np.testing.assert_allclose(J, J2)
    np.testing.assert_allclose(w, w2)

    # test setting with highly non-injective plans
    n = 6
    a = ot.unif(n)
    b_list = [a] * 3
    pi_list = [a[:, None] @ a[None, :]] * 3
    J, w, log_dict = ot.lp.NorthWestMMGluing(pi_list, log=True)
    # Test the validity of the gluing
    gamma = log_dict["gamma"]
    verify_gluing_validity(gamma, J, w, pi_list)


@pytest.skip_backend("tf")  # skips because of array assignment
@pytest.skip_backend("jax")
def test_north_west_mm_gluing_backends(nx):
    rng = np.random.RandomState(0)
    n = 7
    nk_list = [5, 6, 4]
    a = rng.rand(n)
    a = a / np.sum(a)
    b_list = [rng.rand(nk) for nk in nk_list]
    b_list = [b / np.sum(b) for b in b_list]
    M_list = [rng.rand(n, nk) for nk in nk_list]
    pi_list = [ot.emd(a, b, M) for b, M in zip(b_list, M_list)]

    pi_list2 = [nx.from_numpy(pi) for pi in pi_list]
    J, w, log_dict = ot.lp.NorthWestMMGluing(pi_list2, log=True, nx=nx)
    gamma = log_dict["gamma"]

    # Test equality with numpy solution
    J_np, w_np, log_dict_np = ot.lp.NorthWestMMGluing(pi_list, log=True)
    gamma_np = log_dict_np["gamma"]
    np.testing.assert_allclose(J, J_np)
    np.testing.assert_allclose(w, w_np)
    np.testing.assert_allclose(gamma, gamma_np)


def test_clean_discrete_measure(nx):
    a = nx.ones(3) / 3.0
    X = nx.from_numpy(np.array([[1.0, 1.0], [1.0, 1.0], [2.0, 2.0]]))
    X_clean, a_clean = ot.lp._barycenter_solvers._clean_discrete_measure(X, a)
    a_true = nx.from_numpy(np.array([2 / 3, 1 / 3]))
    X_true = nx.from_numpy(np.array([[1.0, 1.0], [2.0, 2.0]]))
    assert a_clean.shape == a_true.shape
    assert X_clean.shape == X_true.shape
    np.testing.assert_allclose(a_clean, a_true)
    np.testing.assert_allclose(X_clean, X_true)

    a = nx.ones(3) / 3.0
    X = nx.from_numpy(np.array([[1.0, 1.0], [2.0, 2.0], [1.0, 1.0]]))
    X_clean, a_clean = ot.lp._barycenter_solvers._clean_discrete_measure(X, a, nx=nx)
    a_true = nx.from_numpy(np.array([2 / 3, 1 / 3]))
    X_true = nx.from_numpy(np.array([[1.0, 1.0], [2.0, 2.0]]))
    assert a_clean.shape == a_true.shape
    assert X_clean.shape == X_true.shape
    np.testing.assert_allclose(a_clean, a_true)
    np.testing.assert_allclose(X_clean, X_true)

    n = 5
    a = nx.ones(n) / n
    v = nx.from_numpy(np.array([1.0, 2.0, 3.0]))
    X = nx.stack([v] * n, axis=0)
    X_clean, a_clean = ot.lp._barycenter_solvers._clean_discrete_measure(X, a)
    a_true = np.array([1.0])
    X_true = np.array([1.0, 2.0, 3.0]).reshape(1, 3)
    assert a_clean.shape == a_true.shape
    assert X_clean.shape == X_true.shape
    np.testing.assert_allclose(a_clean, a_true)
    np.testing.assert_allclose(X_clean, X_true)


def test_to_int_array(nx):
    a_np = np.array([1.0, 2.0, 3.0])
    a = nx.from_numpy(a_np)
    a_int = ot.lp._barycenter_solvers._to_int_array(a)
    a_np_int = a_np.astype(int)
    np.testing.assert_allclose(nx.to_numpy(a_int), a_np_int)
    ot.lp._barycenter_solvers._to_int_array(a, nx=nx)


@pytest.mark.skipif(not ot.lp._barycenter_solvers.cvxopt, reason="No cvxopt available")
def test_lp_barycenter_cvxopt():
    a1 = np.array([1.0, 0, 0])[:, None]
    a2 = np.array([0, 0, 1.0])[:, None]

    A = np.hstack((a1, a2))
    M = np.array([[0, 1.0, 4.0], [1.0, 0, 1.0], [4.0, 1.0, 0]])

    # obvious barycenter between two Diracs
    bary0 = np.array([0, 1.0, 0])

    bary = ot.lp.barycenter(A, M, [0.5, 0.5], solver=None)

    np.testing.assert_allclose(bary, bary0, rtol=1e-5, atol=1e-7)
    np.testing.assert_allclose(bary.sum(), 1)


def test_warnings():
    n = 100  # nb bins
    m = 100  # nb bins

    mean1 = 30
    mean2 = 50

    # bin positions
    x = np.arange(n, dtype=np.float64)
    y = np.arange(m, dtype=np.float64)

    # Gaussian distributions
    a = gauss(n, m=mean1, s=5)  # m= mean, s= std

    b = gauss(m, m=mean2, s=10)

    # loss matrix
    M = ot.dist(x.reshape((-1, 1)), y.reshape((-1, 1))) ** (1.0 / 2)

    print("Computing {} EMD ".format(1))
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        print("Computing {} EMD ".format(1))
        ot.emd(a, b, M, numItermax=1)
        assert "numItermax" in str(w[-1].message)
        # assert len(w) == 1


def test_dual_variables():
    n = 500  # nb bins
    m = 600  # nb bins

    mean1 = 300
    mean2 = 400

    # bin positions
    x = np.arange(n, dtype=np.float64)
    y = np.arange(m, dtype=np.float64)

    # Gaussian distributions
    a = gauss(n, m=mean1, s=5)  # m= mean, s= std

    b = gauss(m, m=mean2, s=10)

    # loss matrix
    M = ot.dist(x.reshape((-1, 1)), y.reshape((-1, 1))) ** (1.0 / 2)

    print("Computing {} EMD ".format(1))

    # emd loss 1 proc
    ot.tic()
    G, log = ot.emd(a, b, M, log=True)
    ot.toc("1 proc : {} s")

    ot.tic()
    G2 = ot.emd(b, a, np.ascontiguousarray(M.T))
    ot.toc("1 proc : {} s")

    cost1 = (G * M).sum()
    # Check symmetry
    np.testing.assert_array_almost_equal(cost1, (M * G2.T).sum())
    # Check with closed-form solution for gaussians
    np.testing.assert_almost_equal(cost1, np.abs(mean1 - mean2))

    # Check that both cost computations are equivalent
    np.testing.assert_almost_equal(cost1, log["cost"])
    check_duality_gap(a, b, M, G, log["u"], log["v"], log["cost"])

    constraint_violation = log["u"][:, None] + log["v"][None, :] - M

    assert constraint_violation.max() < 1e-8


def check_duality_gap(a, b, M, G, u, v, cost):
    cost_dual = np.vdot(a, u) + np.vdot(b, v)
    # Check that dual and primal cost are equal
    np.testing.assert_almost_equal(cost_dual, cost)

    [ind1, ind2] = np.nonzero(G)

    # Check that reduced cost is zero on transport arcs
    np.testing.assert_array_almost_equal(
        (M - u.reshape(-1, 1) - v.reshape(1, -1))[ind1, ind2], np.zeros(ind1.size)
    )



"""
Functions for plotting OT matrices

.. warning::
    Note that by default the module is not import in :mod:`ot`. In order to
    use it you need to explicitly import :mod:`ot.plot`


"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#
# License: MIT License

import numpy as np
import matplotlib.pylab as pl
from matplotlib import gridspec


def plot1D_mat(
    a,
    b,
    M,
    title="",
    plot_style="yx",
    a_label="",
    b_label="",
    color_source="b",
    color_target="r",
    coupling_cmap="gray_r",
):
    r"""Plot matrix :math:`\mathbf{M}` with the source and target 1D distributions.

    Creates a subplot with the source distribution :math:`\mathbf{a}` and target
    distribution :math:`\mathbf{b}`.
    In 'yx' mode (default), the source is on the left and
    the target on the top, and in 'xy' mode, source on the bottom (upside
    down) and the target on the left.
    The matrix :math:`\mathbf{M}` is shown in between.

    Parameters
    ----------
    a : ndarray, shape (na,)
        Source distribution
    b : ndarray, shape (nb,)
        Target distribution
    M : ndarray, shape (na, nb)
        Matrix to plot
    title : str, optional
        Title of the plot
    plot_style : str, optional
        Style of the plot, 'yx' or 'xy'. 'yx' places the source on the left and
        the target on the top, 'xy' places the source on the bottom (upside
        down) and the target on the left.
    a_label : str, optional
        Label for source distribution
    b_label : str, optional
        Label for target distribution
    color_source : str, optional
        Color of the source distribution
    color_target : str, optional
        Color of the target distribution
    coupling_cmap : str, optional
        Colormap for the coupling matrix

    Returns
    -------
    ax1 : source plot ax
    ax2 : target plot ax
    ax3 : coupling plot ax

    See Also
    --------
    :func:`rescale_for_imshow_plot`
    """
    assert plot_style in ["yx", "xy"], "plot_style should be 'yx' or 'xy'"
    na, nb = M.shape

    gs = gridspec.GridSpec(
        3, 3, height_ratios=[1, 1, 1], width_ratios=[1, 1, 1], hspace=0, wspace=0
    )

    xa = np.arange(na)
    xb = np.arange(nb)

    # helper function for code factorisation
    def _set_ticks_and_spines(ax, empty_ticks=True, visible_spines=False):
        if empty_ticks:
            ax.set_xticks(())
            ax.set_yticks(())

        ax.spines["top"].set_visible(visible_spines)
        ax.spines["right"].set_visible(visible_spines)
        ax.spines["bottom"].set_visible(visible_spines)
        ax.spines["left"].set_visible(visible_spines)

    if plot_style == "xy":
        # horizontal source on the bottom, flipped vertically
        ax1 = pl.subplot(gs[2, 1:])
        ax1.plot(xa, np.max(a) - a, color=color_source, linewidth=2)
        ax1.fill(
            xa,
            np.max(a) - a,
            np.max(a) * np.ones_like(a),
            color=color_source,
            alpha=0.5,
        )
        ax1.set_title(a_label, y=-0.15)

        # vertical target on the left
        ax2 = pl.subplot(gs[0:2, 0])
        ax2.plot(b, xb, color=color_target, linewidth=2)
        ax2.fill(b, xb, color=color_target, alpha=0.5)
        ax2.invert_xaxis()
        ax2.invert_yaxis()
        ax2.set_title(b_label)

        _set_ticks_and_spines(ax1, empty_ticks=True, visible_spines=False)
        _set_ticks_and_spines(ax2, empty_ticks=True, visible_spines=False)

        # coupling matrix in the middle
        ax3 = pl.subplot(gs[0:2, 1:], sharey=ax2, sharex=ax1)
        ax3.imshow(M.T, interpolation="nearest", origin="lower", cmap=coupling_cmap)
        ax3.set_title(title)
        _set_ticks_and_spines(ax3, empty_ticks=False, visible_spines=True)

        pl.subplots_adjust(hspace=0, wspace=0)
        return ax1, ax2, ax3

    else:  # plot_style == 'yx'
        # vertical source on the left
        ax1 = pl.subplot(gs[1:, 0])
        ax1.plot(a, xa, color=color_source, linewidth=2)
        ax1.fill(a, xa, color=color_source, alpha=0.5)
        ax1.invert_xaxis()
        ax1.set_title(a_label)

        # horizontal target on the top
        ax2 = pl.subplot(gs[0, 1:])
        ax2.plot(xb, b, color=color_target, linewidth=2)
        ax2.fill(xb, b, color=color_target, alpha=0.5)
        ax2.set_title(b_label)

        _set_ticks_and_spines(ax1, empty_ticks=True, visible_spines=False)
        _set_ticks_and_spines(ax2, empty_ticks=True, visible_spines=False)

        # coupling matrix in the middle
        ax3 = pl.subplot(gs[1:, 1:], sharey=ax1, sharex=ax2)
        ax3.imshow(M, interpolation="nearest", cmap=coupling_cmap)
        # Set title below matrix plot
        ax3.text(
            0.5,
            -0.025,
            title,
            ha="center",
            va="top",
            transform=ax3.transAxes,
            fontsize="large",
        )
        _set_ticks_and_spines(ax3, empty_ticks=False, visible_spines=True)

        pl.subplots_adjust(hspace=0, wspace=0)
        return ax1, ax2, ax3


def rescale_for_imshow_plot(x, y, n, m=None, a_y=None, b_y=None):
    r"""
    Gives arrays xr, yr that can be plotted over an (n, m)
    imshow plot (in 'xy' coordinates). If `a_y` or `b_y` is provided,
    y is sliced over its indices such that y stays in [ay, by].

    Parameters
    ----------
    x : ndarray, shape (nx,)
    y : ndarray, shape (nx,)
    n : int
        x-axis size of the imshow plot on which to plot (x, y)
    m : int, optional
        y-axis size of the imshow plot, defaults to n
    a_y : float, optional
        Lower bound for y
    b_y : float, optional
        Upper bound for y

    Returns
    -------
    xr : ndarray, shape (nx,)
        Rescaled x values (due to slicing, may have less elements than x)
    yr : ndarray, shape (nx,)
        Rescaled y values (due to slicing, may have less elements than y)

    See Also
    --------
    :func:`plot1D_mat`

    """
    # slice over the y values that are in the y range
    a_x, b_x = np.min(x), np.max(x)
    assert x.shape[0] == y.shape[0], "x and y arrays should have the same size"
    if a_y is None:
        a_y = np.min(y)
    if b_y is None:
        b_y = np.max(y)
    if m is None:
        m = n
    idx = (y >= a_y) & (y <= b_y)
    x_rescaled = (x[idx] - a_x) * (n - 1) / (b_x - a_x)
    y_rescaled = (y[idx] - a_y) * (m - 1) / (b_y - a_y)
    return x_rescaled, y_rescaled


def plot2D_samples_mat(xs, xt, G, thr=1e-8, **kwargs):
    r"""Plot matrix :math:`\mathbf{G}` in 2D with lines using alpha values

    Plot lines between source and target 2D samples with a color
    proportional to the value of the matrix :math:`\mathbf{G}` between samples.


    Parameters
    ----------
    xs : ndarray, shape (ns,2)
        Source samples positions
    b : ndarray, shape (nt,2)
        Target samples positions
    G : ndarray, shape (na,nb)
        OT matrix
    thr : float, optional
        threshold above which the line is drawn
    **kwargs : dict
        parameters given to the plot functions (default color is black if
        nothing given)
    """

    if ("color" not in kwargs) and ("c" not in kwargs):
        kwargs["color"] = "k"
    mx = G.max()
    if "alpha" in kwargs:
        scale = kwargs["alpha"]
        del kwargs["alpha"]
    else:
        scale = 1
    for i in range(xs.shape[0]):
        for j in range(xt.shape[0]):
            if G[i, j] / mx > thr:
                pl.plot(
                    [xs[i, 0], xt[j, 0]],
                    [xs[i, 1], xt[j, 1]],
                    alpha=G[i, j] / mx * scale,
                    **kwargs,
                )



# distutils: language = c++
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

import numpy as np
cimport numpy as np
cimport cython
import heapq


@cython.boundscheck(False)
@cython.wraparound(False)
def insert_new_chain(np.ndarray[np.int64_t, ndim=1] chains_starting_at, np.ndarray[np.int64_t, ndim=1] chains_ending_at, int i, int j):
    """Insert the `candidate_chain=(i,j)` into the already known chains stored in 
    `chains_starting_at` and `chains_ending_at`.
    `chains_starting_at` and `chains_ending_at` are modified in-place and 
    the chain in which the candidate is inserted is returned.
    """
    cdef int n = chains_starting_at.shape[0]
    if i - 1 >= 0 and i - 1 < n and chains_ending_at[i - 1] != -1:
        i = chains_ending_at[i - 1]
    if j + 1 >= 0 and j + 1 < n and chains_starting_at[j + 1] != -1:
        j = chains_starting_at[j + 1]
    if i >= 0 and i < n:
        if chains_starting_at[i] != -1:
            chains_ending_at[chains_starting_at[i]] = -1
        chains_starting_at[i] = j
    if j >= 0 and j < n:
        if chains_ending_at[j] != -1:
            chains_starting_at[chains_ending_at[j]] = -1
        chains_ending_at[j] = i
    return i, j

@cython.boundscheck(False)
@cython.wraparound(False)
def compute_cost_for_chain(int idx_start, int idx_end, np.ndarray[np.float64_t, ndim=1] chain_costs_cumsum):
    """Compute the associated cost for a chain (set of contiguous points
    included in the solution) ranging from `idx_start` to `idx_end` (both included).
    """
    if idx_start == 0:
        return chain_costs_cumsum[idx_end]
    else:
        return chain_costs_cumsum[idx_end] - chain_costs_cumsum[idx_start - 1]

@cython.boundscheck(False)
@cython.wraparound(False)
def precompute_chain_costs_cumsum(np.ndarray[np.int64_t, ndim=1] minimal_chain_ending_at_idx, np.ndarray[np.float64_t, ndim=1] minimal_chain_ending_at_cost, int n):
    """For each position `i` at which a chain could end,
    Compute (using dynamic programming and the costs of minimal chains that have been precomputed)
    the cost of the largest chain ending at `i`.

    This is useful because this can be used later, to compute the cost of any chain in O(1)
    (cf. `compute_cost_for_chain`).
    """
    cdef np.ndarray[np.float64_t, ndim=1] chain_costs_cumsum = np.zeros((n, ), dtype=np.float64)
    cdef int i, start
    cdef double additional_cost
    for i in range(n):
        if minimal_chain_ending_at_idx[i] != -1:
            start = minimal_chain_ending_at_idx[i]
            assert start >= 0 and start < n, "start should be in [0, n-1]"
            additional_cost = minimal_chain_ending_at_cost[i]
            if start == 0:
                chain_costs_cumsum[i] = additional_cost
            else:
                chain_costs_cumsum[i] = chain_costs_cumsum[start - 1] + additional_cost
    return chain_costs_cumsum

@cython.boundscheck(False)
@cython.wraparound(False)
def get_cost_w1(np.ndarray[np.float64_t, ndim=1] diff_cum_sum, int idx_start, int idx_end):
    """Compute L1 cost of a chain using the L1 specific trick (O(1) time)."""
    if idx_start == 0:
        return diff_cum_sum[idx_end]
    else:
        return diff_cum_sum[idx_end] - diff_cum_sum[idx_start - 1]

@cython.boundscheck(False)
@cython.wraparound(False)
def get_cost_wp(np.ndarray[np.float64_t, ndim=1] sorted_z, np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator, int idx_start, int idx_end, int p):
    """Compute Lp^p cost of a chain using no specific trick (O(N) time)."""
    cdef np.ndarray[np.float64_t, ndim=1] subset_z = sorted_z[idx_start:idx_end+1]
    cdef np.ndarray[np.int64_t, ndim=1] subset_indicator = sorted_distrib_indicator[idx_start:idx_end+1]
    cdef np.ndarray[np.float64_t, ndim=1] subset_x = subset_z[subset_indicator == 0]
    cdef np.ndarray[np.float64_t, ndim=1] subset_y = subset_z[subset_indicator == 1]
    return np.sum(np.abs(subset_x - subset_y) ** p)


@cython.boundscheck(False)
@cython.wraparound(True)
def compute_costs(np.ndarray[np.float64_t, ndim=1] sorted_z,
                 np.ndarray[np.float64_t, ndim=1] diff_cum_sum,
                 np.ndarray[np.int64_t, ndim=1] diff_ranks,
                 np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator,
                 int p=1):
    """For each element in sorted `z`, compute its minimal chain (cf note below).
    Then compute the cost for each minimal chain and sort all minimal chains in increasing 
    cost order.

    Note: the "minimal chain" of a point x_i in x is the minimal set of adjacent 
    points (starting at x_i and extending to the right) that one should 
    take to get a balanced set (ie. a set in which we have as many 
    elements from x as elements from y)
    """
    cdef list l_costs = []
    cdef np.ndarray[np.int64_t, ndim=1] minimal_chain_ending_at_idx = np.full((sorted_z.shape[0], ), -1, dtype=np.int64)
    cdef np.ndarray[np.float64_t, ndim=1] minimal_chain_ending_at_cost = np.full((sorted_z.shape[0], ), -1.0, dtype=np.float64)
    cdef np.ndarray[np.int64_t, ndim=1] last_pos_for_rank_x = np.full((2 * sorted_z.shape[0], ), -1, dtype=np.int64)
    cdef np.ndarray[np.int64_t, ndim=1] last_pos_for_rank_y = np.full((2 * sorted_z.shape[0], ), -1, dtype=np.int64)
    cdef int n = diff_ranks.shape[0]
    cdef int idx_end, cur_rank, idx_start, target_rank
    cdef double cost
    assert sorted_z.shape[0] == diff_ranks.shape[0] == sorted_distrib_indicator.shape[0], \
        "All input arrays must have the same length"
    for idx_end in range(n):
        # For each item in either distrib, find the scope of the smallest
        # "minimal chain" that would end at that point and extend on the left, 
        # if one exists, and store the cost of this "minimal chain" by relying 
        # on differences of cumulative sums
        cur_rank = diff_ranks[idx_end]
        idx_start = -1
        if sorted_distrib_indicator[idx_end] == 0:
            target_rank = cur_rank - 1
            if last_pos_for_rank_y[target_rank] != -1:
                idx_start = last_pos_for_rank_y[target_rank]
            last_pos_for_rank_x[cur_rank] = idx_end
        else:
            target_rank = cur_rank + 1
            if last_pos_for_rank_x[target_rank] != -1:
                idx_start = last_pos_for_rank_x[target_rank]
            last_pos_for_rank_y[cur_rank] = idx_end
        if idx_start != -1:
            if p == 1:
                cost = get_cost_w1(diff_cum_sum, idx_start, idx_end)
            else:
                cost = get_cost_wp(sorted_z, sorted_distrib_indicator, idx_start, idx_end, p)
            if idx_end == idx_start + 1:
                heapq.heappush(l_costs, (abs(cost), idx_start, idx_end))
            minimal_chain_ending_at_idx[idx_end] = idx_start
            minimal_chain_ending_at_cost[idx_end] = abs(cost)
    return l_costs, precompute_chain_costs_cumsum(minimal_chain_ending_at_idx, minimal_chain_ending_at_cost, n)

@cython.boundscheck(False)
@cython.wraparound(False)
def arg_insert_in_sorted(np.ndarray[np.float64_t, ndim=1] x, np.ndarray[np.float64_t, ndim=1] y):
    """
    Returns the index (similar to argsort) to be used to sort 
    the concatenation of `x` and `y` 
    (supposed to be 1d arrays).

    `x` and `y` are supposed to be sorted and their
    order cannot be changed in the resulting array (which is important
    for our ranking based algo in case of ex-aequos).
    """
    assert x.ndim == 1 and y.ndim == 1
    cdef Py_ssize_t n_x = x.shape[0]
    cdef Py_ssize_t n_y = y.shape[0]
    cdef np.ndarray[np.int64_t, ndim=1] arr_out = np.zeros(n_x + n_y, dtype=np.int64)
    cdef Py_ssize_t idx_x = 0
    cdef Py_ssize_t idx_y = 0
    while idx_x + idx_y < n_x + n_y:
        if idx_x == n_x:
            arr_out[idx_x + idx_y] = n_x + idx_y
            idx_y += 1
        elif idx_y == n_y:
            arr_out[idx_x + idx_y] = idx_x
            idx_x += 1
        else:
            if x[idx_x] <= y[idx_y]:
                arr_out[idx_x + idx_y] = idx_x
                idx_x += 1
            else:
                arr_out[idx_x + idx_y] = n_x + idx_y
                idx_y += 1
    return arr_out


@cython.boundscheck(False)
@cython.wraparound(False)
def preprocess(np.ndarray[np.float64_t, ndim=1] x, np.ndarray[np.float64_t, ndim=1] y):
    """Given two 1d distributions `x` and `y`:
    1. `indices_sort_x` sorts `x` (ie. `x[indices_sort_x]` is sorted) and
       `indices_sort_y` sorts `y` (ie. `y[indices_sort_y]` is sorted)
    2. stack them into a single distrib such that:
    * the new distrib is sorted with sort indices (wrt a stack of sorted x and sorted y) `indices_sort_xy`
    * `sorted_distrib_indicator` is a vector of zeros and ones where 0 means 
        "this point comes from x" and 1 means "this point comes from y"
    """
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_x = np.argsort(x)
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_y = np.argsort(y)
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_xy = arg_insert_in_sorted(x[indices_sort_x], y[indices_sort_y])
    cdef np.ndarray[np.int64_t, ndim=1] idx = np.concatenate((np.zeros(x.shape[0], dtype=np.int64), np.ones(y.shape[0], dtype=np.int64)))
    cdef np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator = idx[indices_sort_xy]
    return indices_sort_x, indices_sort_y, indices_sort_xy, sorted_distrib_indicator

@cython.boundscheck(False)
@cython.wraparound(False)
def generate_solution_using_marginal_costs(
    list costs,
    np.ndarray[np.int64_t, ndim=1] ranks_xy,
    np.ndarray[np.float64_t, ndim=1] chain_costs_cumsum,
    int max_iter,
    np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
):
    """Generate a solution from a sorted list of minimal chain costs.
    See the note in `compute_costs` docs for a definition of minimal chains.

    The solution is a pair of lists. The first list contains the indices from `sorted_x`
    that are in the active set, and the second one contains the indices from `sorted_y`
    that are in the active set. 
    The third returned element is a list of marginal costs induced by each step:
    `arr_marginal_costs[i]` is the marginal cost induced by the `i`-th step of the algorithm, such
    that `np.cumsum(arr_marginal_costs[:i])` gives all intermediate costs up to step `i`.
    """
    cdef np.ndarray[np.int64_t, ndim=1] active_set = np.zeros((sorted_distrib_indicator.shape[0], ), dtype=np.int64)
    cdef np.ndarray[np.int64_t, ndim=1] chains_starting_at = np.full((sorted_distrib_indicator.shape[0], ), -1, dtype=np.int64)
    cdef np.ndarray[np.int64_t, ndim=1] chains_ending_at = np.full((sorted_distrib_indicator.shape[0], ), -1, dtype=np.int64)
    cdef np.ndarray[np.int64_t, ndim=2] active_set_inserts = np.zeros((max_iter, 2), dtype=np.int64)
    cdef int n = sorted_distrib_indicator.shape[0]
    cdef int i, j, p_s, p_e
    cdef int n_pairs_in_active_set = 0
    cdef double c
    cdef double marginal_cost
    cdef np.ndarray[np.float64_t, ndim=1] arr_marginal_costs = np.zeros((max_iter, ), dtype=np.float64)

    while len(costs) > 0 and max_iter > n_pairs_in_active_set:
        c, i, j = heapq.heappop(costs)
        if active_set[i] == 1 or active_set[j] == 1:
            continue
        # Case 1: j == i + 1 => "Simple" insert
        if j == i + 1:
            p_s, p_e = insert_new_chain(chains_starting_at, chains_ending_at, i, j)
        # Case 2: insert a chain that contains a chain
        elif chains_ending_at[j - 1] != -1:
            if i + 1 >= 0 and i + 1 < n:
                chains_starting_at[i + 1] = -1
            if j - 1 >= 0 and j - 1 < n:
                chains_ending_at[j - 1] = -1
            p_s, p_e = insert_new_chain(chains_starting_at, chains_ending_at, i, j)
        # There should be no "Case 3"
        else:
            raise ValueError
        active_set_inserts[n_pairs_in_active_set, 0] = i
        active_set_inserts[n_pairs_in_active_set, 1] = j
        arr_marginal_costs[n_pairs_in_active_set] = c
        active_set[i] = 1
        active_set[j] = 1
        n_pairs_in_active_set += 1
        
        # We now need to update the candidate chains wrt the chain we have just created
        if p_s == 0 or p_e == n - 1:
            continue
        if sorted_distrib_indicator[p_s - 1] != sorted_distrib_indicator[p_e + 1]:
            # Insert (p_s - 1, p_e + 1) as a new candidate chain with marginal cost
            marginal_cost = (compute_cost_for_chain(p_s - 1, p_e + 1, chain_costs_cumsum)
                             - compute_cost_for_chain(p_s, p_e, chain_costs_cumsum))
            heapq.heappush(costs, (marginal_cost, p_s - 1, p_e + 1))
    # Generate index arrays in the order of insertion in the active set
    cdef np.ndarray[np.int64_t, ndim=1] indices_sorted_x = np.zeros((max_iter, ), dtype=np.int64) - 1
    cdef np.ndarray[np.int64_t, ndim=1] indices_sorted_y = np.zeros((max_iter, ), dtype=np.int64) - 1
    cdef int a, b
    for i in range(max_iter):
        a = active_set_inserts[i, 0]
        b = active_set_inserts[i, 1]
        if sorted_distrib_indicator[a] == 0:
            indices_sorted_x[i] = ranks_xy[a]
        else:
            indices_sorted_y[i] = ranks_xy[a]
        if sorted_distrib_indicator[b] == 0:
            indices_sorted_x[i] = ranks_xy[b]
        else:
            indices_sorted_y[i] = ranks_xy[b]
    return (
        indices_sorted_x,
        indices_sorted_y,
        arr_marginal_costs
    )

@cython.boundscheck(False)
@cython.wraparound(False)
def compute_cumulative_sum_differences(
    np.ndarray[np.float64_t, ndim=1] x_sorted, 
    np.ndarray[np.float64_t, ndim=1] y_sorted, 
    np.ndarray[np.int64_t, ndim=1] indices_sort_xy, 
    np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
):
    """Computes difference between cumulative sums for both distribs.

    The cumulative sum vector for a sorted x is:

        cumsum_x = [x_0, x_0 + x_1, ..., x_0 + ... + x_n]

    This vector is then extend to reach a length of 2*n by repeating 
    values at places that correspond to an y item.
    In other words, if the order of x and y elements on the real 
    line is something like x-y-y-x..., then the extended vector is 
    (note the repetitions):

        cumsum_x = [x_0, x_0, x_0, x_0 + x_1, ..., x_0 + ... + x_n]

    Overall, this function returns `cumsum_x - cumsum_y` where `cumsum_x`
    and `cumsum_y` are the extended versions.
    """
    cdef np.ndarray[np.float64_t, ndim=1] cum_sum_xs = np.cumsum(x_sorted)
    cdef np.ndarray[np.float64_t, ndim=1] cum_sum_ys = np.cumsum(y_sorted)
    cdef np.ndarray[np.float64_t, ndim=1] cum_sum = np.concatenate((cum_sum_xs, cum_sum_ys))
    cdef np.ndarray[np.float64_t, ndim=1] cum_sum_ordered = cum_sum[indices_sort_xy]

    cdef np.ndarray[np.float64_t, ndim=1] cum_sum_x = insert_constant_values_float(cum_sum_ordered, 0, sorted_distrib_indicator)
    cdef np.ndarray[np.float64_t, ndim=1] cum_sum_y = insert_constant_values_float(cum_sum_ordered, 1, sorted_distrib_indicator)

    return cum_sum_x - cum_sum_y

@cython.boundscheck(False)
@cython.wraparound(False)
def insert_constant_values_float(
    np.ndarray[np.float64_t, ndim=1] arr, 
    np.int64_t distrib_index,
    np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
):
    """Takes `arr` as input. For each position `i` in `arr`, 
    if `sorted_distrib_indicator[i]==distrib_index`,
    the value from `arr` is copied, 
    otherwise, the previous value that was copied from `arr` is repeated.
    """
    cdef np.ndarray[np.float64_t, ndim=1] arr_insert = np.copy(arr)
    for i in range(len(arr)):
        if sorted_distrib_indicator[i]!=distrib_index:
            if i == 0:
                arr_insert[i] = 0
            else:
                arr_insert[i] = arr_insert[i-1]
    return arr_insert

@cython.boundscheck(False)
@cython.wraparound(False)
def insert_constant_values_int(
    np.ndarray[np.int64_t, ndim=1] arr, 
    np.int64_t distrib_index,
    np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
):
    """Takes `arr` as input. For each position `i` in `arr`, 
    if `sorted_distrib_indicator[i]==distrib_index`,
    the value from `arr` is copied, 
    otherwise, the previous value that was copied from `arr` is repeated.
    """
    cdef np.ndarray[np.int64_t, ndim=1] arr_insert = np.copy(arr)
    for i in range(len(arr)):
        if sorted_distrib_indicator[i]!=distrib_index:
            if i == 0:
                arr_insert[i] = 0
            else:
                arr_insert[i] = arr_insert[i-1]
    return arr_insert

@cython.boundscheck(False)
@cython.wraparound(False)
def compute_rank_differences(
    np.ndarray[np.int64_t, ndim=1] indices_sort_xy, 
    np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
):
    """Precompute important rank-related quantities for better minimal chain extraction.
    
    Two quantities are returned:

    * `ranks_xy` is an array that gathers ranks of the elements in 
        their original distrib, eg. if the distrib indicator is
        [0, 1, 1, 0, 0, 1], then `rank_xy` will be:
        [0, 0, 1, 1, 2, 2]
    * `diff_ranks` is computed from `ranks_xy_x_cum` and `ranks_xy_y_cum`.
        For the example above, we would have:
        
        ranks_xy_x_cum = [1, 1, 1, 2, 3, 3]
        ranks_xy_y_cum = [0, 1, 2, 2, 2, 3]

        And `diff_ranks` is just `ranks_xy_x_cum - ranks_xy_y_cum`.
    """
    cdef int n_x = np.sum(sorted_distrib_indicator == 0)
    cdef int n_y = np.sum(sorted_distrib_indicator == 1)
    cdef np.ndarray[np.int64_t, ndim=1] ranks_x = np.arange(n_x)
    cdef np.ndarray[np.int64_t, ndim=1] ranks_y = np.arange(n_y)
    cdef np.ndarray[np.int64_t, ndim=1]ranks_xy = np.concatenate((ranks_x, ranks_y))
    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy_ordered = ranks_xy[indices_sort_xy]

    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy_x = ranks_xy_ordered.copy()
    ranks_xy_x[sorted_distrib_indicator==1] = 0
    ranks_xy_x[sorted_distrib_indicator==0] += 1
    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy_x_cum = insert_constant_values_int(ranks_xy_x, 0, sorted_distrib_indicator)

    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy_y = ranks_xy_ordered.copy()
    ranks_xy_y[sorted_distrib_indicator==0] = 0
    ranks_xy_y[sorted_distrib_indicator==1] += 1
    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy_y_cum = insert_constant_values_int(ranks_xy_y, 1, sorted_distrib_indicator)

    cdef np.ndarray[np.int64_t, ndim=1] diff_ranks = ranks_xy_x_cum - ranks_xy_y_cum

    return ranks_xy_ordered, diff_ranks

@cython.boundscheck(False)
@cython.wraparound(False)
def partial_wasserstein_1d_cy(
    np.ndarray[np.float64_t, ndim=1] x, 
    np.ndarray[np.float64_t, ndim=1] y, 
    int max_iter, 
    int p
):
    """Main routine for the partial Wasserstein problem in 1D.
    
    Does:
    
    1. Preprocessing of the distribs (sorted & co)
    2. Precomputations (ranks, cumulative sums)
    3. Extraction of minimal chains
    4. Generate and return solution

    Note that the indices in `indices_x` and `indices_y` are ordered wrt their order of
    appearance in the solution such that `indices_x[:10]` (resp y) is the set of indices
    from x (resp. y) for the partial problem of size 10.

    Arguments
    ---------
    x : np.ndarray of shape (n, )
        First distrib to be considered (weights are considered uniform)
    y : np.ndarray of shape (m, )
        Second distrib to be considered (weights are considered uniform)
    max_iter : int
        Number of iterations of the algorithm, which is equal to the number of pairs
        in the returned solution.
    p : int (default: 1)
        Order of the partial Wasserstein distance to be computed (p-Wasserstein, or $W_p^p$)

    Returns
    -------
    indices_x : np.ndarray of shape (min(n, m, max_iter), )
        Indices of elements from the x distribution to be included in the partial solutions
        Order of appearance in this array indicates order of inclusion in the solution
    indices_y : np.ndarray of shape (min(n, m, max_iter), )
        Indices of elements from the x distribution to be included in the partial solutions
        Order of appearance in this array indicates order of inclusion in the solution
    list_marginal_costs : list of length min(n, m, max_iter)
        List of marginal costs associated to the intermediate partial problems
        `np.cumsum(list_marginal_costs)` gives the corresponding total costs for intermediate partial problems
    """
    # Sort distribs and keep track of their original indices
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_x
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_y
    cdef np.ndarray[np.int64_t, ndim=1] indices_sort_xy
    cdef np.ndarray[np.int64_t, ndim=1] sorted_distrib_indicator
    cdef np.ndarray[np.float64_t, ndim=1] sorted_z
    cdef np.ndarray[np.float64_t, ndim=1] diff_cum_sum
    cdef np.ndarray[np.int64_t, ndim=1] ranks_xy
    cdef np.ndarray[np.int64_t, ndim=1] diff_ranks
    cdef list costs
    cdef np.ndarray[np.float64_t, ndim=1] chain_costs_cumsum
    cdef np.ndarray[np.int64_t, ndim=1] indices_x
    cdef np.ndarray[np.int64_t, ndim=1] indices_y
    cdef np.ndarray[np.int64_t, ndim=1] sol_indices_x_sorted
    cdef np.ndarray[np.int64_t, ndim=1] sol_indices_y_sorted
    cdef np.ndarray[np.float64_t, ndim=1] marginal_costs

    indices_sort_x, indices_sort_y, indices_sort_xy, sorted_distrib_indicator = preprocess(x, y)

    sorted_z = np.concatenate((x[indices_sort_x], y[indices_sort_y]))[indices_sort_xy]

    # Precompute useful quantities
    diff_cum_sum = compute_cumulative_sum_differences(x[indices_sort_x], 
                                                      y[indices_sort_y], 
                                                      indices_sort_xy,
                                                      sorted_distrib_indicator)
    ranks_xy, diff_ranks = compute_rank_differences(indices_sort_xy, sorted_distrib_indicator)

    # Compute costs for "minimal chains"
    costs, chain_costs_cumsum = compute_costs(sorted_z, diff_cum_sum, diff_ranks, sorted_distrib_indicator, p=p)

    # Generate solution from sorted costs
    sol_indices_x_sorted, sol_indices_y_sorted, marginal_costs = generate_solution_using_marginal_costs(costs, 
                                                                                                   ranks_xy, 
                                                                                                   chain_costs_cumsum, 
                                                                                                   max_iter, 
                                                                                                   sorted_distrib_indicator)

    # Convert back into indices in original `x` and `y` distribs
    indices_x = indices_sort_x[sol_indices_x_sorted]
    indices_y = indices_sort_y[sol_indices_y_sorted]
    return indices_x, indices_y, marginal_costs



# -*- coding: utf-8 -*-
"""
Batch operations for quadratic optimal transport.
"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#         Paul Krzakala <paul.krzakala@gmail.com>
#
# License: MIT License

from ..utils import OTResult
from ot.backend import get_backend
from ot.batch._linear import loss_linear_batch
from ot.batch._utils import bmv, bop, bregman_log_projection_batch


def tensor_batch(
    a, b, C1, C2, symmetric=True, nx=None, loss="sqeuclidean", logits=None
):
    r"""
    Compute the Gromov-Wasserstein cost tensor for a batch of problems.

    The Gromov-Wasserstein distance can be expressed as:

    .. math::
        \text{GW}(\mathbf{T}, \mathbf{C}_1, \mathbf{C}_2) = \sum_{ijkl} T_{ik} T_{jl} \ell(C_{1,ij}, C_{2,kl}) = \langle \mathcal{L} \times \mathbf{T}, \mathbf{T} \rangle

    where :math:`\mathcal{L}` is a 4D tensor with elements :math:`\mathcal{L}[i,j,k,l] = \ell(C_{1,ij}, C_{2,kl})`.

    For loss functions of the form :math:`\ell(a,b) = f_1(a) + f_2(b) - \langle h_1(a), h_2(b) \rangle`,
    the tensor product :math:`\mathcal{L} \times \mathbf{T}` can be computed efficiently without explicitly computing :math:`\mathcal{L}` [12].

    This function precomputes all matrices that implicitly define `\mathcal{L}` for various loss functions.

    Parameters
    ----------
    a : array-like, shape (B, n)
        Source distributions for each problem in the batch.
    b : array-like, shape (B, m)
        Target distributions for each problem in the batch.
    C1 : array-like, shape (B, n, n) or (B, n, n, d)
        Source cost matrices for each problem. Can be a 3D array for scalar costs or a 4D array for vector-valued costs (edge features).
    C2 : array-like, shape (B, m, m) or (B, n, n, d)
        Target cost matrices for each problem. Can be a 3D array for scalar costs or a 4D array for vector-valued costs (edge features).
    symmetric : bool, optional
        Whether the cost matrices are symmetric. Default is True.
    nx : backend object, optional
        Numerical backend to use for computations. If None, the default backend is used.
    loss : str, optional
        Loss function to use. Supported values: 'sqeuclidean', 'kl'.
        Default is 'sqeuclidean'.
    logits : bool, optional
        For KL divergence, whether inputs are logits (unnormalized log probabilities).
        If True, inputs are treated as logits. Default is None.

    Returns
    -------
    dict
        Dictionary containing:
        - constC : array-like, shape (B, n, m)
            Constant term in the tensor product.
        - hC1 : array-like, shape (B, n, n, d) or (B, n, n)
        - hC2 : array-like, shape (B, m, m, d) or (B, m, m)
        - fC1 : array-like, shape (B, n, n)
        - fC2 : array-like, shape (B, m, m)


    Supported loss functions:
    ------------------------------

    **Squared Euclidean loss**:

    .. math::
        \ell(a, b) = \|a - b\|_2^2 = \sum_i (a_i - b_i)^2

    **KL divergence**:

    .. math::
        \ell(a, b) = \sum_i a_i \log\left(\frac{a_i}{b_i}\right)

    If ``logits=True``, the entries of C1 are treated as logits (unnormalized log probabilities)
    and the loss becomes:

    .. math::
        \ell(x, y) = \sum_i y_i (\log(y_i) - x_i)

    Examples
    --------
    >>> import numpy as np
    >>> from ot.batch import tensor_batch
    >>> # Create batch of cost matrices
    >>> C1 = np.random.rand(3, 5, 5)  # 3 problems, 5x5 source matrices
    >>> C2 = np.random.rand(3, 4, 4)  # 3 problems, 4x4 target matrices
    >>> a = np.ones((3, 5)) / 5  # Uniform source distributions
    >>> b = np.ones((3, 4)) / 4  # Uniform target distributions
    >>> L = tensor_batch(a, b, C1, C2, loss='sqeuclidean')

    References
    ----------
    .. [12] Gabriel Peyré, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.
    """

    if nx is None:
        nx = get_backend(C1)

    if loss == "sqeuclidean":

        def f1(C1):
            if C1.ndim == 4:
                return nx.sum(C1**2, axis=-1)
            else:
                return C1**2

        def f2(C2):
            if C2.ndim == 4:
                return nx.sum(C2**2, axis=-1)
            else:
                return C2**2

        def h1(C1):
            if C1.ndim == 3:
                C1 = nx.unsqueeze(C1, -1)
            return 2 * C1

        def h2(C2):
            if C2.ndim == 3:
                C2 = nx.unsqueeze(C2, -1)
            return C2

    elif loss == "kl":
        assert logits in [
            True,
            False,
        ], "logits must be either True or False for KL loss"

        def f1(C1):
            return nx.zeros((C1.shape[0], C1.shape[1], C1.shape[2]), type_as=C1)

        def f2(C2):
            assert C2.ndim == 4, "C2 must be a bxnxnxd tensor"
            fC2 = C2 * nx.log(C2 + 1e-15)  # Avoid log(0)
            return nx.sum(fC2, axis=-1)

        def h1(C1):
            return C1 if logits else nx.log(C1 + 1e-15)

        def h2(C2):
            return C2

    return compute_tensor_batch(f1, f2, h1, h2, a, b, C1, C2, symmetric=symmetric)


def loss_quadratic_batch(L, T, recompute_const=False, symmetric=True, nx=None):
    r"""
    Computes the gromov-wasserstein cost given a cost tensor and transport plan. Batched version.

    Parameters
    ----------
    L : dict
        Cost tensor as returned by `tensor_batch`.
    T : array-like, shape (B, n, m)
        Transport plan.
    recompute_const : bool, optional
        Whether to recompute the constant term. Default is False. This should be set to True if T does not satisfy the marginal constraints.
    symmetric : bool, optional
        Whether to use symmetric version. Default is True.
    nx : module, optional
        Backend to use. Default is None.

    Examples
    --------
    >>> import numpy as np
    >>> from ot.batch import tensor_batch, loss_quadratic_batch
    >>> # Create batch of cost matrices
    >>> C1 = np.random.rand(3, 5, 5)  # 3 problems, 5x5 source matrices
    >>> C2 = np.random.rand(3, 4, 4)  # 3 problems, 4x4 target matrices
    >>> a = np.ones((3, 5)) / 5  # Uniform source distributions
    >>> b = np.ones((3, 4)) / 4  # Uniform target distributions
    >>> L = tensor_batch(a, b, C1, C2, loss='sqeuclidean')
    >>> # Use the uniform transport plan for testing
    >>> T = np.ones((3, 5, 4)) / (5 * 4)
    >>> loss = loss_quadratic_batch(L, T, recompute_const=True)
    >>> loss.shape
    (3,)

    See Also
    --------
    ot.batch.tensor_batch : From computing the cost tensor L.
    ot.batch.solve_gromov_batch : For finding the optimal transport plan T.
    """
    if nx is None:
        nx = get_backend(T)
    LT = tensor_product_batch(
        L, T, nx=nx, recompute_const=recompute_const, symmetric=symmetric
    )
    return nx.sum(LT * T, axis=(1, 2))


def loss_quadratic_samples_batch(
    a,
    b,
    C1,
    C2,
    T,
    loss="sqeuclidean",
    symmetric=None,
    nx=None,
    logits=None,
    recompute_const=False,
):
    r"""
    Computes the gromov-wasserstein for samples C1, C2 and transport plan. Batched version.

    Parameters
    ----------
    a : array-like, shape (B, n)
        Source distributions.
    b : array-like, shape (B, m)
        Target distributions.
    C1 : array-like, shape (B, n, n) or (B, n, n, d)
        Source cost matrices.
    C2 : array-like, shape (B, m, m) or (B, n, n, d)
        Target cost matrices.
    T : array-like, shape (B, n, m)
        Transport plan.
    loss : str, optional
        Loss function to use. Supported values: 'sqeuclidean', 'kl'.
        Default is 'sqeuclidean'.
    recompute_const : bool, optional
        Whether to recompute the constant term. Default is False. This should be set to True if T does not satisfy the marginal constraints.
    symmetric : bool, optional
        Whether to use symmetric version. Default is True.
    nx : module, optional
        Backend to use. Default is None.

    Examples
    --------
    >>> import numpy as np
    >>> from ot.batch import loss_quadratic_samples_batch
    >>> # Create batch of cost matrices
    >>> C1 = np.random.rand(3, 5, 5)  # 3 problems, 5x5 source matrices
    >>> C2 = np.random.rand(3, 4, 4)  # 3 problems, 4x4 target matrices
    >>> a = np.ones((3, 5)) / 5  # Uniform source distributions
    >>> b = np.ones((3, 4)) / 4  # Uniform target distributions
    >>> # Use the uniform transport plan for testing
    >>> T = np.ones((3, 5, 4)) / (5 * 4)
    >>> loss = loss_quadratic_samples_batch(a, b, C1, C2, T, recompute_const=True)
    >>> loss.shape
    (3,)

    See Also
    --------
    ot.batch.tensor_batch : From computing the cost tensor L.
    ot.batch.solve_gromov_batch : For finding the optimal transport plan T.
    """
    if isinstance(loss, str):
        L = tensor_batch(
            a, b, C1, C2, symmetric=symmetric, nx=nx, loss=loss, logits=logits
        )
    else:
        raise ValueError(f"Unknown loss function: {loss}")
    return loss_quadratic_batch(
        L, T, recompute_const=recompute_const, symmetric=symmetric, nx=nx
    )


def solve_gromov_batch(
    C1,
    C2,
    reg=1e-2,
    a=None,
    b=None,
    loss="sqeuclidean",
    symmetric=None,
    M=None,
    alpha=None,
    T_init=None,
    max_iter=50,
    tol=1e-5,
    max_iter_inner=50,
    tol_inner=1e-5,
    grad="envelope",
    logits=None,
):
    r"""
    Solves a batch of Gromov-Wasserstein optimal transport problems using proximal gradient [12, 81].
    For each problem in the batch, solves:

    .. math::
        \begin{aligned}
            \min_{\mathbf{T} \geq 0} \quad & \sum_{i,j,k,l} L(\mathbf{C_1}_{i,k}, \mathbf{C_2}_{j,l}) \mathbf{T}_{i,j} \mathbf{T}_{k,l} \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    If :math:`\mathbf{M}` and :math:`\alpha` are given, solves the more general fused Gromov-Wasserstein problem:

    .. math::
        \begin{aligned}
            \min_{\mathbf{T} \geq 0} \quad & (1-\alpha) \sum_{i,j} M_{i,j} \mathbf{T}_{i,j} + \alpha \sum_{i,j,k,l} L(\mathbf{C_1}_{i,k}, \mathbf{C_2}_{j,l}) \mathbf{T}_{i,j} \mathbf{T}_{k,l} \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    Writing the objective as :math:`(1-\alpha) \langle \mathbf{M}, \mathbf{T} \rangle + \alpha \langle \mathcal{L} \otimes \mathbf{T}, \mathbf{T} \rangle`,
    the solver uses proximal gradient descent where each iteration is:

    .. math::
        \begin{aligned}
            \mathbf{T}_{k+1} = \mathop{\arg \min}_{\mathbf{T} \geq 0} \quad & \langle \mathbf{M}_k, \mathbf{T} \rangle + \epsilon \, \text{KL}(\mathbf{T} \| \mathbf{T}_k) \\
            \text{where} \quad & \mathbf{M}_k = (1 - \alpha) \mathbf{M} + 2 \alpha \mathcal{L} \otimes \mathbf{T}_k
        \end{aligned}

    This can be rewritten as:

    .. math::
        \begin{aligned}
            \mathbf{T}_{k+1} = \mathop{\arg \min}_{\mathbf{T} \geq 0} \quad & \langle \mathbf{M}_k - \epsilon \log(\mathbf{T}_k), \mathbf{T} \rangle - \epsilon H(\mathbf{T})
        \end{aligned}

    where :math:`H(\mathbf{T})` is the entropy of :math:`\mathbf{T}`. Thus each iteration can be solved using the Bregman projection solver implemented in `bregman_log_projection_batch`. 
    
    Note that the inner optimization problem does not need to be solved exactly. In practice it sufficient to set `max_iter_inner` to a low value (e.g. 20) and/or `tol_inner` to a high value (e.g. 1e-2).

    Parameters
    ----------
    C1 : array-like, shape (B, n, n, d) or (B, n, n)
        Samples affinity matrices from source distribution
    C2 : array-like, shape (B, n, n, d) or (B, n, n)
        Samples affinity matrices from target distribution
    a : array-like, shape (B, n), optional
        Marginal distribution of the source samples. If None, uniform distribution is used.
    b : array-like, shape (B, m), optional
        Marginal distribution of the target samples. If None, uniform distribution is used.
    loss : str, optional
        Type of loss function, can be 'sqeuclidean' or 'kl' or a QuadraticMetric instance.
    symmetric : bool, optional
        Either C1 and C2 are to be assumed symmetric or not.
        If let to its default None value, a symmetry test will be conducted.
        Else if set to True (resp. False), C1 and C2 will be assumed symmetric (resp. asymmetric).
    M : array-like, shape (dim_a, dim_b), optional
        Linear cost matrix for Fused Gromov-Wasserstein (default is None).
    alpha : float, optional
        Weight the quadratic term (alpha*Gromov) and the linear term
        ((1-alpha)*Wass) in the Fused Gromov-Wasserstein problem. Not used for
        Gromov problem (when M is not provided). By default ``alpha=None``
        corresponds to ``alpha=1`` for Gromov problem (``M==None``) and
        ``alpha=0.5`` for Fused Gromov-Wasserstein problem (``M!=None``)
    epsilon : float, optional
        Regularization parameter for proximal gradient descent. Default is 1e-2.
    T_init : array-like, shape (B, n, m), optional
        Initial transport plan. If None, it is initialized to uniform distribution.
    max_iter : int, optional
        Maximum number of iterations for the proximal gradient descent. Default is 50.
    tol : float, optional
        Tolerance for convergence of the proximal gradient descent. Default is 1e-5.
    max_iter_inner : int, optional
        Maximum number of iterations for the inner Bregman projection. Default is 50.
    tol_inner : float, optional
        Tolerance for convergence of the inner Bregman projection. Default is 1e-5.
    grad : str, optional
        Type of gradient computation, either or 'autodiff', 'envelope' or 'detach'. 'autodiff' provides gradients wrt all
        outputs (`plan, value, value_linear`) but with important memory cost.
        'envelope' provides gradients only for (`value, value_linear`)`.  `detach`` is the fastest option but
        provides no gradients. Default is 'detach'.
    assume_inner_convergence : bool, optional
        If True, assumes that the inner Bregman projection always converged i.e. the transport plan satisfies the marginal constraints.
        This enables faster computations of the tensor product but might results in inaccurate results (e.g. negative values of the loss).
        Default is True.
    Returns
    -------
    res : OTResult()
        Result of the optimization problem. The information can be obtained as follows:

        - res.plan : OT plan :math:`\mathbf{T}`
        - res.potentials : OT dual potentials
        - res.value : Optimal value of the optimization problem
        - res.value_linear : Linear OT loss with the optimal OT plan
        - res.value_quad : Quadratic OT loss with the optimal OT plan

        See :any:`OTResult` for more information.

    See Also
    --------
    ot.batch.tensor_batch : From computing the cost tensor L.
    ot.solve_gromov : Non-batched solver for Gromov-Wasserstein. Note that the non-batched solver uses a different algorithm (conditional gradient) and might not give the same results.
    
    References
    ----------
    .. [12] Gabriel Peyré, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.
    .. [81] Xu, H., Luo, D., & Carin, L. (2019). "Scalable Gromov-Wasserstein learning for graph partitioning and matching."
        Advances in neural information processing systems (NeurIPS). 2019.
    """

    # -------------- Setup -------------- #

    nx = get_backend(a, b, M, C1, C2, T_init)
    B, n, m = (C1.shape[0], C1.shape[1], C2.shape[1])

    if a is None:
        a = nx.ones((B, n), type_as=C1) / n
    if b is None:
        b = nx.ones((B, m), type_as=C2) / m

    if symmetric is None:
        symmetric = nx.allclose(C1, transpose(C1, nx=nx), atol=1e-10) and nx.allclose(
            C2, transpose(C2, nx=nx), atol=1e-10
        )

    # -------------- Get cost_tensor (quadratic part) -------------- #

    if isinstance(loss, str):
        L = tensor_batch(
            a, b, C1, C2, symmetric=symmetric, nx=nx, loss=loss, logits=logits
        )
    else:
        raise ValueError(f"Unknown loss function: {loss}")

    # -------------- Get cost_matrix (linear part) -------------- #

    if M is None and alpha is None:
        M = nx.zeros((B, n, m), type_as=C1)
        alpha = 1.0  # Gromov problem
    elif M is not None and alpha is None:
        raise ValueError(
            "If M is provided, alpha must also be provided for Fused Gromov-Wasserstein problem"
        )
    elif M is None and alpha is not None:
        raise ValueError(
            "If alpha is provided, M must also be provided for Fused Gromov-Wasserstein problem"
        )
    elif M is not None and alpha is not None:
        if M.shape[0] != B or M.shape[1] != n or M.shape[2] != m:
            raise ValueError(
                f"Shape of M {M.shape} does not match the batch size {B} and dimensions {n}, {m}"
            )
        if alpha < 0 or alpha > 1:
            raise ValueError(
                "Alpha must be in [0, 1] for Fused Gromov-Wasserstein problem"
            )

    # -------------- Solver -------------- #

    if T_init is None:
        const = nx.sqrt(nx.sum(a, axis=1) * nx.sum(b, axis=1))
        T_init = bop(a, b, nx=nx) / const[:, None, None]  # B x n x m

    grad_inner = "autodiff" if grad == "autodiff" else "detach"

    T = T_init
    log_T = nx.log(T + 1e-15)  # Avoid log(0)
    for iter in range(max_iter):
        T_prev = T
        Mk = (1 - alpha) * M + 2 * alpha * tensor_product_batch(
            L, T, nx=nx, recompute_const=False, symmetric=symmetric
        )
        K = -Mk / reg + log_T
        out = bregman_log_projection_batch(
            K, a, b, nx=nx, max_iter=max_iter_inner, tol=tol_inner, grad=grad_inner
        )
        T = out["T"]
        log_T = out["log_T"]
        max_err = nx.max(nx.mean(nx.abs(T_prev - T), axis=(1, 2)))
        if max_err < tol:
            break

    if grad == "detach":
        T = nx.detach(T)
        M = nx.detach(M)
        L = detach_cost_tensor(L, nx=nx)
    elif grad == "envelope":
        T = nx.detach(T)

    value_linear = loss_linear_batch(M, T, nx=nx)
    value_quadratic = loss_quadratic_batch(
        L, T, nx=nx, recompute_const=True, symmetric=symmetric
    )  # Always recompute const for accurate value
    value = (1 - alpha) * value_linear + alpha * value_quadratic
    log = {"n_iter": iter}

    res = OTResult(
        value=value,
        value_linear=value_linear
        * (1 - alpha),  # Weight the linear value for consistency with ot.solve_gromov
        value_quad=value_quadratic * alpha,  # idem
        plan=T,
        backend=nx,
        potentials=out["potentials"],
        log=log,
    )

    return res


### --------------------- Utility functions for quadratic OT --------------------- ###


def compute_tensor_batch(f1, f2, h1, h2, a, b, C1, C2, symmetric=True):
    """
    Gromov-Wasserstein writes as:
        GW(T,C1,C2) = sum_ijkl T_ik T_jl l(C1_ij, C2_kl) = < LxT, T >
    Where L is a cost tensor L[i,j,k,l] = l(C1_ij, C2_kl).

    For loss function of form l(a,b) = f1(a) + f2(b) - < h1(a), h2(b) >
    The tensor product LxT can be computed fast using tensor_product [12].

    References
    ----------
    .. [12] Gabriel Peyré, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.
    """

    fC1 = f1(C1)
    fC2 = f2(C2)
    if not symmetric:
        fC1 = 0.5 * (fC1 + transpose(fC1))
        fC2 = 0.5 * (fC2 + transpose(fC2))
    hC1 = h1(C1)
    hC2 = h2(C2)

    constC = compute_const_from_marginals(fC1, fC2, a, b)

    L = {"constC": constC, "hC1": hC1, "hC2": hC2, "fC1": fC1, "fC2": fC2}

    return L


def tensor_product_batch(L, T, nx=None, recompute_const=False, symmetric=True):
    """
    Compute the tensor product LxT for the cost tensor L and transport plan T.
    The formula is:
        LxT = const - hC1 T hC2^T
        const = < fC1 a 1^T + 1 (fC2 b)^T

    References
    ----------
    .. [12] Gabriel Peyré, Marco Cuturi, and Justin Solomon,
        "Gromov-Wasserstein averaging of kernel and distance matrices."
        International Conference on Machine Learning (ICML). 2016.
    """

    if nx is None:
        nx = get_backend(T)

    if recompute_const:
        const = compute_const_from_marginals(
            L["fC1"], L["fC2"], nx.sum(T, axis=2), nx.sum(T, axis=1), nx=nx
        )
    else:
        const = L["constC"]

    hC1 = L["hC1"]
    hC2 = L["hC2"]

    dot = nx.einsum("bijd,bjk->bikd", hC1, T)
    dot = nx.einsum("bikd,bjkd->bijd", dot, hC2)
    dot = nx.sum(dot, axis=-1)

    if not symmetric:
        dot_t = nx.einsum("bijd,bjk->bikd", transpose(hC1), T)
        dot_t = nx.einsum("bikd,bjkd->bijd", dot_t, transpose(hC2))
        dot_t = nx.sum(dot_t, axis=-1)
        dot = (dot + dot_t) / 2  # Average the two symmetric terms

    return const - dot


def transpose(C, nx=None):
    if nx is None:
        nx = get_backend(C)
    return nx.transpose(C, (0, 2, 1)) if C.ndim == 3 else nx.transpose(C, (0, 2, 1, 3))


def compute_const_from_marginals(fC1, fC2, a, b, nx=None):
    """
    Compute the constant term f1(C1) a 1^T + 1 b^T f2(C2)^T
    """
    if nx is None:
        nx = get_backend(fC1, fC2, a, b)
    fC1a = bmv(fC1, a, nx=nx)
    fC2b = bmv(fC2, b, nx=nx)
    constC = fC1a[:, :, None] + fC2b[:, None, :]
    return constC


def detach_cost_tensor(L, nx=None):
    """
    Detach the cost tensor L to avoid gradients.
    """
    if nx is None:
        nx = get_backend(L["constC"], L["hC1"], L["hC2"])
    L_detached = {}
    for key, value in L.items():
        L_detached[key] = nx.detach(value)
    return L_detached



# -*- coding: utf-8 -*-
"""
Solvers for the original linear program OT problem.

"""

# Author: Remi Flamary <remi.flamary@unice.fr>
#
# License: MIT License

from .dmmot import dmmot_monge_1dgrid_loss, dmmot_monge_1dgrid_optimize
from ._network_simplex import emd, emd2
from ._barycenter_solvers import (
    barycenter,
    free_support_barycenter,
    generalized_free_support_barycenter,
    free_support_barycenter_generic_costs,
    ot_barycenter_energy,
    NorthWestMMGluing,
)
from ..utils import check_number_threads

# import compiled emd
from .emd_wrap import emd_1d_sorted
from .solver_1d import (
    emd_1d,
    emd2_1d,
    wasserstein_1d,
    binary_search_circle,
    wasserstein_circle,
    semidiscrete_wasserstein2_unif_circle,
    linear_circular_ot,
)

__all__ = [
    "emd",
    "emd2",
    "barycenter",
    "free_support_barycenter",
    "cvx",
    "emd_1d_sorted",
    "emd_1d",
    "emd2_1d",
    "wasserstein_1d",
    "generalized_free_support_barycenter",
    "binary_search_circle",
    "wasserstein_circle",
    "semidiscrete_wasserstein2_unif_circle",
    "linear_circular_ot",
    "dmmot_monge_1dgrid_loss",
    "dmmot_monge_1dgrid_optimize",
    "check_number_threads",
    "free_support_barycenter_generic_costs",
    "NorthWestMMGluing",
    "ot_barycenter_energy",
]
